# -*- coding: utf-8 -*-

from collections import defaultdict, OrderedDict
from itertools import permutations
import math

import pytest

try:
    import numpy as np
except ImportError:
    np = None


from chempy import Equilibrium, Reaction, ReactionSystem, Substance
from chempy.thermodynamics.expressions import MassActionEq
from chempy.units import (
    SI_base_registry,
    get_derived_unit,
    allclose,
    units_library,
    linspace,
    to_unitless,
    default_constants as const,
    default_units as u,
)
from chempy.util._expr import Expr
from chempy.util.testing import requires
from .test_rates import _get_SpecialFraction_rsys
from ..arrhenius import ArrheniusParam
from ..rates import Arrhenius, MassAction, Radiolytic, RampedTemp
from .._rates import ShiftedTPoly
from ..ode import (
    get_odesys,
    chained_parameter_variation,
    _mk_dedim,
    _create_odesys as create_odesys,
)
from ..integrated import dimerization_irrev, binary_rev


@requires("numpy", "pyodesys")
def test_get_odesys_1():
    k = 0.2
    a = Substance("A")
    b = Substance("B")
    r = Reaction({"A": 1}, {"B": 1}, param=k)
    rsys = ReactionSystem([r], [a, b])
    assert sorted(rsys.substances.keys()) == ["A", "B"]
    odesys = get_odesys(rsys, include_params=True)[0]
    c0 = {
        "A": 1.0,
        "B": 3.0,
    }
    t = np.linspace(0.0, 10.0)
    xout, yout, info = odesys.integrate(t, c0)
    yref = np.zeros((t.size, 2))
    yref[:, 0] = np.exp(-k * t)
    yref[:, 1] = 4 - np.exp(-k * t)
    assert np.allclose(yout, yref)


@requires("numpy", "pyodesys", "sympy")
def test_get_odesys__rate_exprs_cb():
    k = 0.2
    a = Substance("A")
    b = Substance("B")
    r = Reaction({"A": 1}, {"B": 1}, param=k)
    rsys = ReactionSystem([r], [a, b])
    assert sorted(rsys.substances.keys()) == ["A", "B"]
    odesys, extra = get_odesys(rsys)
    c0 = {"A": 1.0, "B": 3.0}
    t = np.linspace(0.0, 10.0)
    res = odesys.integrate(t, c0)
    yref = np.zeros((t.size, 2))
    yref[:, 0] = np.exp(-k * t)
    yref[:, 1] = 4 - np.exp(-k * t)
    assert np.allclose(res.yout, yref)
    rate = extra["rate_exprs_cb"](res.xout, res.yout, res.params)
    assert np.allclose(rate[:, 0], k * yref[:, 0])


@requires("numpy", "pyodesys")
def test_get_odesys_2():
    g = Radiolytic([3.14])
    a = Substance("A")
    b = Substance("B")
    r = Reaction({"A": 1}, {"B": 1}, param=g)
    rsys = ReactionSystem([r], [a, b])
    odesys = get_odesys(rsys, include_params=True)[0]
    c0 = {
        "A": 1.0,
        "B": 3.0,
    }
    t = np.linspace(0.0, 0.1)
    xout, yout, info = odesys.integrate(
        t, rsys.as_per_substance_array(c0), {"doserate": 2.72, "density": 0.998}
    )
    yref = np.zeros((t.size, 2))
    k = 3.14 * 2.72 * 0.998
    yref[:, 0] = 1 - k * t
    yref[:, 1] = 3 + k * t
    assert np.allclose(yout, yref)


@requires(units_library, "pyodesys")
def test_get_odesys_3():
    M = u.molar
    s = u.second
    mol = u.mol
    m = u.metre
    substances = list(map(Substance, "H2O H+ OH-".split()))
    dissociation = Reaction({"H2O": 1}, {"H+": 1, "OH-": 1}, 2.47e-5 / s)
    recombination = Reaction({"H+": 1, "OH-": 1}, {"H2O": 1}, 1.37e11 / M / s)
    rsys = ReactionSystem([dissociation, recombination], substances)
    odesys = get_odesys(
        rsys, include_params=True, unit_registry=SI_base_registry, output_conc_unit=M
    )[0]
    c0 = {"H2O": 55.4 * M, "H+": 1e-7 * M, "OH-": 1e-4 * mol / m ** 3}
    x, y, p = odesys.to_arrays(
        -42 * u.second, rsys.as_per_substance_array(c0, unit=M), ()
    )
    fout = odesys.f_cb(x, y, p)

    time_unit = get_derived_unit(SI_base_registry, "time")
    conc_unit = get_derived_unit(SI_base_registry, "concentration")

    r1 = to_unitless(55.4 * 2.47e-5 * M / s, conc_unit / time_unit)
    r2 = to_unitless(1e-14 * 1.37e11 * M / s, conc_unit / time_unit)
    assert np.all(abs(fout[:, 0] - r2 + r1)) < 1e-10
    assert np.all(abs(fout[:, 1] - r1 + r2)) < 1e-10
    assert np.all(abs(fout[:, 2] - r1 + r2)) < 1e-10


@requires(units_library, "pyodesys")
def test_get_odesys__with_units():
    a = Substance("A")
    b = Substance("B")
    molar = u.molar
    second = u.second
    r = Reaction({"A": 2}, {"B": 1}, param=1e-3 / molar / second)
    rsys = ReactionSystem([r], [a, b])
    odesys = get_odesys(rsys, include_params=True, unit_registry=SI_base_registry)[0]
    c0 = {"A": 13 * u.mol / u.metre ** 3, "B": 0.2 * u.molar}
    conc_unit = get_derived_unit(SI_base_registry, "concentration")
    t = np.linspace(0, 10) * u.hour
    xout, yout, info = odesys.integrate(
        t, rsys.as_per_substance_array(c0, unit=conc_unit), atol=1e-10, rtol=1e-12
    )

    t_unitless = to_unitless(xout, u.second)
    Aref = dimerization_irrev(t_unitless, 1e-6, 13.0)
    # Aref = 1/(1/13 + 2*1e-6*t_unitless)
    yref = np.zeros((xout.size, 2))
    yref[:, 0] = Aref
    yref[:, 1] = 200 + (13 - Aref) / 2
    assert allclose(yout, yref * conc_unit)


@requires("numpy", "pyodesys")
def test_SpecialFraction():
    k, kprime = 3.142, 2.718
    rsys = _get_SpecialFraction_rsys(k, kprime)

    odesys = get_odesys(rsys, include_params=True)[0]
    c0 = {"H2": 13, "Br2": 17, "HBr": 19}
    r = k * c0["H2"] * c0["Br2"] ** (3 / 2) / (c0["Br2"] + kprime * c0["HBr"])
    ref = rsys.as_per_substance_array({"H2": -r, "Br2": -r, "HBr": 2 * r})
    res = odesys.f_cb(0, rsys.as_per_substance_array(c0))
    assert np.allclose(res, ref)


@requires(units_library, "pyodesys")
def test_SpecialFraction_with_units():
    k, kprime = 3.142 * u.s ** -1 * u.molar ** -0.5, 2.718
    rsys = _get_SpecialFraction_rsys(k, kprime)
    odesys = get_odesys(rsys, include_params=True, unit_registry=SI_base_registry)[0]
    c0 = {"H2": 13 * u.molar, "Br2": 16 * u.molar, "HBr": 19 * u.molar}
    r = k * c0["H2"] * c0["Br2"] ** (3 / 2) / (c0["Br2"] + kprime * c0["HBr"])
    conc_unit = u.mol / u.metre ** 3
    rate_unit = conc_unit / u.second
    ref = rsys.as_per_substance_array(
        {"H2": -r, "Br2": -r, "HBr": 2 * r}, unit=rate_unit
    )
    res = odesys.f_cb(0, rsys.as_per_substance_array(c0, unit=conc_unit))
    assert allclose(to_unitless(ref, rate_unit), res)


@requires("pyodesys")
def test_ode_with_global_parameters():
    ratex = MassAction(Arrhenius([1e10, 40e3 / 8.3145]))
    rxn = Reaction({"A": 1}, {"B": 1}, ratex)
    rsys = ReactionSystem([rxn], "A B")
    odesys, extra = get_odesys(rsys, include_params=False)
    param_keys, unique_keys, p_units = map(
        extra.get, "param_keys unique p_units".split()
    )
    conc = {"A": 3, "B": 5}
    x, y, p = odesys.to_arrays(-37, conc, {"temperature": 298.15})
    fout = odesys.f_cb(x, y, p)
    ref = 3 * 1e10 * np.exp(-40e3 / 8.3145 / 298.15)
    assert np.all(abs((fout[:, 0] + ref) / ref) < 1e-14)
    assert np.all(abs((fout[:, 1] - ref) / ref) < 1e-14)


@requires("pyodesys")
def test_get_ode__ArrheniusParam():
    rxn = Reaction({"A": 1}, {"B": 1}, None)
    rxn.param = ArrheniusParam(1e10, 40e3)
    rsys = ReactionSystem([rxn], "A B")
    odesys = get_odesys(rsys, include_params=True)[0]
    conc = {"A": 3, "B": 5}
    x, y, p = odesys.to_arrays(-37, conc, {"temperature": 200})
    fout = odesys.f_cb(x, y, p)
    ref = 3 * 1e10 * np.exp(-40e3 / 8.314472 / 200)
    assert np.all(abs((fout[:, 0] + ref) / ref) < 1e-14)
    assert np.all(abs((fout[:, 1] - ref) / ref) < 1e-14)


@requires("pyodesys")
def test_get_ode__Radiolytic():
    rad = Radiolytic([2.4e-7])
    rxn = Reaction({"A": 4, "B": 1}, {"C": 3, "D": 2}, rad)
    rsys = ReactionSystem([rxn], "A B C D")
    odesys = get_odesys(rsys, include_params=True)[0]
    c = {"A": 3, "B": 5, "C": 11, "D": 13}
    x, y, p = odesys.to_arrays(-37, c, {"doserate": 0.4, "density": 0.998})
    fout = odesys.f_cb(x, y, p)
    r = 2.4e-7 * 0.4 * 0.998
    ref = [-4 * r, -r, 3 * r, 2 * r]
    assert np.all(abs((fout - ref) / ref) < 1e-14)


@requires("pyodesys", units_library)
def test_get_ode__Radiolytic__units():
    rad = Radiolytic([2.4e-7 * u.mol / u.joule])
    rxn = Reaction({"A": 4, "B": 1}, {"C": 3, "D": 2}, rad)
    rsys = ReactionSystem([rxn], "A B C D")
    odesys = get_odesys(rsys, include_params=True, unit_registry=SI_base_registry)[0]
    conc = {"A": 3 * u.molar, "B": 5 * u.molar, "C": 11 * u.molar, "D": 13 * u.molar}
    x, y, p = odesys.to_arrays(
        -37 * u.second,
        conc,
        {
            "doserate": 0.4 * u.gray / u.second,
            "density": 0.998 * u.kg / u.decimetre ** 3,
        },
    )
    fout = odesys.f_cb(x, y, p)  # f_cb does not carry any units
    r = 2.4e-7 * 0.4 * 0.998 * 1e3  # mol/m3
    ref = [-4 * r, -r, 3 * r, 2 * r]
    assert np.all(abs((fout - ref) / ref) < 1e-14)


@requires("pyodesys", units_library)
def test_get_ode__Radiolytic__units__multi():
    rad = Radiolytic([2.4e-7 * u.mol / u.joule])
    rxn = Reaction({"A": 4, "B": 1}, {"C": 3, "D": 2}, rad)
    rsys = ReactionSystem([rxn], "A B C D")
    odesys = get_odesys(rsys, include_params=True, unit_registry=SI_base_registry)[0]
    conc = {"A": 3 * u.molar, "B": 5 * u.molar, "C": 11 * u.molar, "D": 13 * u.molar}
    doserates = [dr * u.gray / u.second for dr in [0.1, 0.2, 0.3, 0.4]]
    results = odesys.integrate(
        37 * u.second,
        conc,
        {"doserate": doserates, "density": 0.998 * u.kg / u.decimetre ** 3},
    )
    assert len(results) == 4
    for i, r in enumerate(results):
        dr = r.params[odesys.param_names.index("doserate")]
        assert dr.ndim == 0 or len(dr) == 1
        assert dr == doserates[i]


class Density(Expr):
    """Arguments: rho0 drhodT T0"""

    parameter_keys = ("temperature",)
    kw = {}

    def __call__(self, variables, backend=None):
        rho0, drhodT, T0 = self.all_args(variables)
        return rho0 + drhodT * (variables["temperature"] - T0)


@requires("pyodesys")
def test_get_ode__Radiolytic__substitutions():
    rad = Radiolytic([2.4e-7])
    rxn = Reaction({"A": 4, "B": 1}, {"C": 3, "D": 2}, rad)
    rsys = ReactionSystem([rxn], "A B C D")
    substance_rho = Density([1, -1e-3, 273.15])
    odesys = get_odesys(
        rsys, include_params=True, substitutions={"density": substance_rho}
    )[0]
    conc = {"A": 3, "B": 5, "C": 11, "D": 13}
    state = {"doserate": 0.4, "temperature": 298.15}
    x, y, p = odesys.to_arrays(-37, conc, state)
    fout = odesys.f_cb(x, y, p)
    r = 2.4e-7 * 0.4 * substance_rho({"temperature": 298.15})
    ref = [-4 * r, -r, 3 * r, 2 * r]
    assert np.all(abs((fout - ref) / ref) < 1e-14)


@requires("pyodesys", units_library)
def test_get_ode__Radiolytic__substitutions__units():
    rad = Radiolytic([2.4e-7 * u.mol / u.joule])
    rxn = Reaction({"A": 4, "B": 1}, {"C": 3, "D": 2}, rad)
    rsys = ReactionSystem([rxn], "A B C D")
    g_dm3 = u.gram / u.decimetre ** 3
    kg_dm3 = u.kg / u.decimetre ** 3
    substance_rho = Density([1 * kg_dm3, -1 * g_dm3 / u.kelvin, 273.15 * u.kelvin])
    odesys = get_odesys(
        rsys,
        include_params=True,
        unit_registry=SI_base_registry,
        substitutions={"density": substance_rho},
    )[0]
    conc = {"A": 3 * u.molar, "B": 5 * u.molar, "C": 11 * u.molar, "D": 13 * u.molar}
    x, y, p = odesys.to_arrays(
        -37 * u.second,
        conc,
        {"doserate": 0.4 * u.gray / u.second, "temperature": 298.15 * u.kelvin},
    )
    fout = odesys.f_cb(x, y, p)
    r = 2.4e-7 * 0.4 * 0.975 * 1e3  # mol/m3/s
    ref = [-4 * r, -r, 3 * r, 2 * r]
    assert np.all(abs((fout - ref) / ref) < 1e-14)


@requires("pyodesys", units_library)
def test_get_ode__TPoly():
    rate = MassAction(
        ShiftedTPoly([273.15 * u.K, 10 / u.molar / u.s, 2 / u.molar / u.s / u.K])
    )
    rxn = Reaction({"A": 1, "B": 1}, {"C": 3, "D": 2}, rate, {"A": 3})
    rsys = ReactionSystem([rxn], "A B C D")
    odesys = get_odesys(rsys, unit_registry=SI_base_registry)[0]
    conc = {"A": 3 * u.molar, "B": 5 * u.molar, "C": 11 * u.molar, "D": 13 * u.molar}
    x, y, p = odesys.to_arrays(-37 * u.second, conc, {"temperature": 298.15 * u.kelvin})
    fout = odesys.f_cb(x, y, p)
    r = 3 * 5 * (10 + 2 * 25) * 1000  # mol/m3/s
    ref = [-4 * r, -r, 3 * r, 2 * r]
    assert np.all(abs((fout - ref) / ref) < 1e-14)


@requires("pyodesys", units_library)
def test_get_odesys__time_dep_rate():
    class RampedRate(Expr):
        argument_names = ("rate_constant", "ramping_rate")

        def __call__(self, variables, reaction, backend=math):
            rate_constant, ramping_rate = self.all_args(variables, backend=backend)
            return rate_constant * ramping_rate * variables["time"]

    rate = MassAction(RampedRate([7, 2]))
    rxn = Reaction({"A": 1}, {"B": 3}, rate)
    rsys = ReactionSystem([rxn], "A B")
    odesys = get_odesys(rsys)[0]
    conc = {"A": 3, "B": 11}
    x, y, p = odesys.to_arrays([5, 13, 17], conc, ())
    fout = odesys.f_cb(x, y, p)
    r = 2 * 7 * 3
    ref = np.array([[-r * 5, -r * 13, -r * 17], [r * 5 * 3, r * 13 * 3, r * 17 * 3]]).T
    assert np.allclose(fout, ref)


@requires("pyodesys", units_library)
def test_get_odesys__time_dep_temperature():
    import sympy as sp

    def refA(t, A0, A, Ea_over_R, T0, dTdt):
        T = T0 + dTdt * t
        d_Ei = sp.Ei(-Ea_over_R / T0).n(100).round(90) - sp.Ei(-Ea_over_R / T).n(
            100
        ).round(90)
        d_Texp = T0 * sp.exp(-Ea_over_R / T0) - T * sp.exp(-Ea_over_R / T)
        return A0 * sp.exp(A / dTdt * (Ea_over_R * d_Ei + d_Texp)).n(30)

    params = A0, A, Ea_over_R, T0, dTdt = [13, 1e10, 56e3 / 8, 273, 2]
    B0 = 11
    rate = MassAction(Arrhenius([A, Ea_over_R]))
    rxn = Reaction({"A": 1}, {"B": 3}, rate)
    rsys = ReactionSystem([rxn], "A B")
    rt = RampedTemp([T0, dTdt], ("init_temp", "ramp_rate"))
    odesys, extra = get_odesys(rsys, False, substitutions={"temperature": rt})
    all_pk, unique, p_units = map(extra.get, "param_keys unique p_units".split())
    conc = {"A": A0, "B": B0}
    tout = [2, 5, 10]

    for ramp_rate in [2, 3, 4]:
        unique["ramp_rate"] = ramp_rate
        xout, yout, info = odesys.integrate(10, conc, unique, atol=1e-10, rtol=1e-12)
        params[-1] = ramp_rate
        Aref = np.array([float(refA(t, *params)) for t in xout])
        # Aref = 1/(1/13 + 2*1e-6*t_unitless)
        yref = np.zeros((xout.size, 2))
        yref[:, 0] = Aref
        yref[:, 1] = B0 + 3 * (A0 - Aref)
        assert allclose(yout, yref)

    unique["ramp_rate"] = 2
    x, y, p = odesys.to_arrays(tout, conc, unique)
    fout = odesys.f_cb(x, y, p)

    def r(t):
        return A * np.exp(-Ea_over_R / (T0 + dTdt * t)) * A0  # refA(t, *params)

    ref = np.array([[-r(2), -r(5), -r(10)], [3 * r(2), 3 * r(5), 3 * r(10)]]).T
    assert np.allclose(fout, ref)


@requires("numpy", "pyodesys")
def test_get_odesys__late_binding():
    def _gibbs(args, T, R, backend, **kwargs):
        H, S = args
        return backend.exp(-(H - T * S) / (R * T))

    def _eyring(args, T, R, k_B, h, backend, **kwargs):
        H, S = args
        return k_B / h * T * backend.exp(-(H - T * S) / (R * T))

    gibbs_pk = ("temperature", "molar_gas_constant")
    eyring_pk = gibbs_pk + ("Boltzmann_constant", "Planck_constant")

    GibbsEC = MassActionEq.from_callback(
        _gibbs, argument_names=("H", "S"), parameter_keys=gibbs_pk
    )
    EyringMA = MassAction.from_callback(
        _eyring, argument_names=("H", "S"), parameter_keys=eyring_pk
    )

    uk_equil = ("He_assoc", "Se_assoc")
    beta = GibbsEC(unique_keys=uk_equil)  # equilibrium parameters

    uk_kinet = ("Ha_assoc", "Sa_assoc")
    bimol_barrier = EyringMA(unique_keys=uk_kinet)  # activation parameters

    eq = Equilibrium({"Fe+3", "SCN-"}, {"FeSCN+2"}, beta)
    rsys = ReactionSystem(eq.as_reactions(kf=bimol_barrier))
    odesys, extra = get_odesys(rsys, include_params=False)
    pk, unique, p_units = map(extra.get, "param_keys unique p_units".split())
    assert sorted(unique) == sorted(uk_equil + uk_kinet)
    assert sorted(pk) == sorted(eyring_pk)


@requires("numpy", "pyodesys")
def test_get_odesys__ScaledSys():
    from pyodesys.symbolic import ScaledSys

    k = 0.2
    a = Substance("A")
    b = Substance("B")
    r = Reaction({"A": 1}, {"B": 1}, param=k)
    rsys = ReactionSystem([r], [a, b])
    assert sorted(rsys.substances.keys()) == ["A", "B"]
    odesys = get_odesys(rsys, include_params=True, SymbolicSys=ScaledSys)[0]
    c0 = {
        "A": 1.0,
        "B": 3.0,
    }
    t = np.linspace(0.0, 10.0)
    xout, yout, info = odesys.integrate(t, c0)
    yref = np.zeros((t.size, 2))
    yref[:, 0] = np.exp(-k * t)
    yref[:, 1] = 4 - np.exp(-k * t)
    assert np.allclose(yout, yref)


@requires("numpy", "pyodesys", "sympy")
def test_get_odesys__max_euler_step_cb():
    rsys = ReactionSystem.from_string(
        "\n".join(["H2O -> H+ + OH-; 1e-4", "OH- + H+ -> H2O; 1e10"])
    )
    odesys, extra = get_odesys(rsys)
    r1 = 1.01e-4
    r2 = 6e-4
    dH2Odt = r2 - r1
    euler_ref = 2e-7 / dH2Odt
    assert (
        abs(
            extra["max_euler_step_cb"](0, {"H2O": 1.01, "H+": 2e-7, "OH-": 3e-7})
            - euler_ref
        )
        / euler_ref
        < 1e-8
    )


@requires("numpy", "pyodesys", "sympy")
@pytest.mark.parametrize("substances", permutations(["H2O", "H+", "OH-"]))
def test_get_odesys__linear_dependencies__preferred(substances):
    rsys = ReactionSystem.from_string(
        "\n".join(["H2O -> H+ + OH-; 1e-4", "OH- + H+ -> H2O; 1e10"]), substances
    )
    assert isinstance(rsys.substances, OrderedDict)
    odesys, extra = get_odesys(rsys)

    af_H2O_H = extra["linear_dependencies"](["H+", "H2O"])
    import sympy

    y0 = {k: sympy.Symbol(k + "0") for k in rsys.substances}
    af_H2O_H(
        None, {odesys[k]: v for k, v in y0.items()}, None, sympy
    )  # ensure idempotent
    exprs_H2O_H = af_H2O_H(None, {odesys[k]: v for k, v in y0.items()}, None, sympy)
    ref_H2O_H = {
        "H2O": y0["H2O"] + y0["OH-"] - odesys["OH-"],  # oxygen
        "H+": 2 * y0["H2O"]
        + y0["H+"]
        + y0["OH-"]
        - odesys["OH-"]
        - 2 * (y0["H2O"] + y0["OH-"] - odesys["OH-"]),  # hydrogen
    }
    for k, v in ref_H2O_H.items():
        assert (exprs_H2O_H[odesys[k]] - v) == 0


@requires("numpy", "pyodesys", "sympy", "pycvodes")
@pytest.mark.parametrize(
    "preferred", [None, ["H+", "OH-"], ["H2O", "H+"], ["H2O", "OH-"]]
)
def test_get_odesys__linear_dependencies__PartiallySolvedSystem(preferred):
    import sympy
    from pyodesys.symbolic import PartiallySolvedSystem

    rsys = ReactionSystem.from_string(
        "\n".join(["H2O -> H+ + OH-; 1e-4", "OH- + H+ -> H2O; 1e10"])
    )
    odesys, extra = get_odesys(rsys)
    c0 = {"H2O": 0, "H+": 2e-7, "OH-": 3e-7}
    h0max = extra["max_euler_step_cb"](0, c0)
    analytic_factory = extra["linear_dependencies"]()
    y0 = {k: sympy.Symbol(k + "0") for k in rsys.substances}
    analytic_factory(None, {odesys[k]: v for k, v in y0.items()}, None, sympy)
    psys = PartiallySolvedSystem(odesys, analytic_factory)
    xout, yout, info = psys.integrate(
        1, c0, atol=1e-12, rtol=1e-10, first_step=h0max * 1e-12, integrator="cvode"
    )
    c_reac = c0["H+"], c0["OH-"]
    H2O_ref = binary_rev(xout, 1e10, 1e-4, c0["H2O"], max(c_reac), min(c_reac))
    assert np.allclose(yout[:, psys.names.index("H2O")], H2O_ref)
    assert np.allclose(yout[:, psys.names.index("H+")], c0["H+"] + c0["H2O"] - H2O_ref)
    assert np.allclose(
        yout[:, psys.names.index("OH-")], c0["OH-"] + c0["H2O"] - H2O_ref
    )


@requires("numpy", "pyodesys", "sympy", "pycvodes")
def test_get_odesys__Equilibrium_as_reactions():
    from chempy import Equilibrium, ReactionSystem

    eq = Equilibrium({"Fe+3", "SCN-"}, {"FeSCN+2"}, 10 ** 2)
    substances = "Fe+3 SCN- FeSCN+2".split()
    rsys = ReactionSystem(eq.as_reactions(kf=3.0), substances)
    odesys, extra = get_odesys(rsys)
    init_conc = {"Fe+3": 1.0, "SCN-": 0.3, "FeSCN+2": 0}
    tout, Cout, info = odesys.integrate(
        5, init_conc, integrator="cvode", atol=1e-11, rtol=1e-12
    )
    cmplx_ref = binary_rev(
        tout, 3, 3.0 / 100, init_conc["FeSCN+2"], init_conc["Fe+3"], init_conc["SCN-"]
    )
    assert np.allclose(Cout[:, 2], cmplx_ref)


@requires("numpy", "pyodesys", "sympy", "pycvodes")
def test_get_odesys__Expr_as_param():
    def _eyring_pe(args, T, backend=math, **kwargs):
        (freq,) = args
        return freq * T

    EyringPreExp = Expr.from_callback(
        _eyring_pe, argument_names=("freq",), parameter_keys=("temperature",)
    )

    def _k(args, T, backend=math, **kwargs):
        A, H, S = args
        return A * backend.exp(-(H - T * S) / (8.314511 * T))

    EyringMA = MassAction.from_callback(
        _k, parameter_keys=("temperature",), argument_names=("Aa", "Ha", "Sa")
    )
    kb_h = 2.08e10
    rxn = Reaction({"A"}, {"B"}, EyringMA(unique_keys=("A_u", "H_u", "S_u")))
    rsys = ReactionSystem([rxn], ["A", "B"])
    odesys, extra = get_odesys(
        rsys, include_params=False, substitutions={"A_u": EyringPreExp(kb_h)}
    )
    y0 = defaultdict(float, {"A": 7.0})
    rt = 293.15
    xout, yout, info = odesys.integrate(
        5,
        y0,
        {"H_u": 117e3, "S_u": 150, "temperature": rt},
        integrator="cvode",
        atol=1e-12,
        rtol=1e-10,
        nsteps=1000,
    )
    kref = kb_h * rt * np.exp(-(117e3 - rt * 150) / (8.314511 * rt))
    ref = y0["A"] * np.exp(-kref * xout)
    assert np.allclose(yout[:, 0], ref)
    assert np.allclose(yout[:, 1], y0["A"] - ref)


@requires("numpy", "pyodesys", "sympy", "pycvodes")
def test_get_odesys__Expr_as_param__unique_as_param():
    def _eyring_pe_coupled(args, T, S, backend=math, **kwargs):
        (freq,) = args
        return freq * T / S

    EyringPreExpCoupled = Expr.from_callback(
        _eyring_pe_coupled,
        argument_names=("freq",),
        parameter_keys=("temperature", "S_u"),
    )

    def _k(args, T, backend=math, **kwargs):
        A, H, S = args
        return A * backend.exp(-(H - T * S) / (8.314511 * T))

    EyringMA = MassAction.from_callback(
        _k, parameter_keys=("temperature",), argument_names=("Aa", "Ha", "Sa")
    )
    kb_h = 2.08e10
    rxn = Reaction({"A"}, {"B"}, EyringMA(unique_keys=("A_u", "H_u", "S_u")))
    rsys = ReactionSystem([rxn], ["A", "B"])
    odesys2, extra2 = get_odesys(
        rsys, include_params=False, substitutions={"A_u": EyringPreExpCoupled(kb_h)}
    )
    y0 = defaultdict(float, {"A": 7.0})
    rt = 293.15
    xout2, yout2, info2 = odesys2.integrate(
        5,
        y0,
        {"H_u": 107e3, "S_u": 150, "temperature": rt},
        integrator="cvode",
        atol=1e-12,
        rtol=1e-10,
        nsteps=1000,
    )
    kref2 = kb_h * rt * np.exp(-(107e3 - rt * 150) / (8.314511 * rt)) / 150
    ref2 = y0["A"] * np.exp(-kref2 * xout2)
    assert np.allclose(yout2[:, 0], ref2)
    assert np.allclose(yout2[:, 1], y0["A"] - ref2)


@requires("pyodesys", "pycvodes")
def test_chained_parameter_variation():
    ratex = MassAction(Arrhenius([1e10, 63e3 / 8.3145]))
    rxn = Reaction({"A": 1}, {"B": 1}, ratex)
    rsys = ReactionSystem([rxn], "A B")
    odesys, extra = get_odesys(rsys, include_params=False)
    param_keys, unique_keys, p_units = map(
        extra.get, "param_keys unique p_units".split()
    )
    conc = {"A": 3.17, "B": 5.03}
    Ts = (294, 304, 317)
    times = [3.1, 2.1, 5.3]
    kw = dict(integrator="cvode", atol=1e-12, rtol=1e-13, first_step=1e-14)
    tout, cout, info = chained_parameter_variation(
        odesys, times, conc, {"temperature": Ts}, {}, integrate_kwargs=kw
    )
    assert len(info["nfev"]) == 3
    assert info["nfev"][0] > 2
    assert info["nfev"][1] > 2
    assert info["nfev"][2] > 2
    assert np.all(np.diff(tout) > 0)
    tout1 = tout[tout <= times[0]]
    tout23 = tout[tout > times[0]]
    tout2 = tout23[tout23 <= times[0] + times[1]]
    tout3 = tout23[tout23 > times[0] + times[1]]

    def _ref(y0, x, T, x0):
        k = 1e10 * np.exp(-63e3 / 8.3145 / T)
        return y0 * np.exp(-k * (x - x0))

    Aref1 = _ref(conc["A"], tout1, Ts[0], tout1[0])
    Bref1 = conc["B"] + conc["A"] - Aref1

    Aref2 = _ref(Aref1[-1], tout2, Ts[1], tout1[-1])
    Bref2 = Bref1[-1] + Aref1[-1] - Aref2

    Aref3 = _ref(Aref2[-1], tout3, Ts[2], tout2[-1])
    Bref3 = Bref2[-1] + Aref2[-1] - Aref3

    cref = np.concatenate(
        [
            np.vstack((a, b)).T
            for a, b in [(Aref1, Bref1), (Aref2, Bref2), (Aref3, Bref3)]
        ]
    )
    forgive = 27 * 1.1
    assert np.allclose(cref, cout, atol=kw["atol"] * forgive, rtol=kw["rtol"] * forgive)


def _check_cstr(odesys, fr, fc, extra_pars=None):
    tout, c0 = np.linspace(0, 0.13, 7), {"H2O2": 2, "O2": 4, "H2O": 3}
    params = {fr: 13, fc["H2O2"]: 11, fc["O2"]: 43, fc["H2O"]: 45}
    params.update(extra_pars or {})
    res = odesys.integrate(tout, c0, params)
    from chempy.kinetics.integrated import binary_irrev_cstr

    def get_analytic(result, k, n):
        ref = binary_irrev_cstr(
            result.xout,
            5,
            result.named_dep("H2O2")[0],
            result.named_dep(k)[0],
            result.named_param(fc["H2O2"]),
            result.named_param(fc[k]),
            result.named_param(fr),
            n,
        )
        return np.array(ref).T

    ref_O2 = get_analytic(res, "O2", 1)
    ref_H2O = get_analytic(res, "H2O", 2)
    assert np.allclose(res.named_dep("H2O2"), ref_O2[:, 0])
    assert np.allclose(res.named_dep("H2O2"), ref_H2O[:, 0])
    assert np.allclose(res.named_dep("O2"), ref_O2[:, 1])
    assert np.allclose(res.named_dep("H2O"), ref_H2O[:, 1])


@requires("pyodesys", "scipy", "sym")
def test_get_odesys__cstr():
    rsys = ReactionSystem.from_string("2 H2O2 -> O2 + 2 H2O; 5")
    odesys, extra = get_odesys(rsys, cstr=True)
    fr, fc = extra["cstr_fr_fc"]
    _check_cstr(odesys, fr, fc)


@requires("pyodesys", "scipy", "sym")
def test_create_odesys__cstr():
    rsys = ReactionSystem.from_string("2 H2O2 -> O2 + 2 H2O; 'k2'")
    fr, fc = "feedratio", OrderedDict([(sk, "fc_%s" % sk) for sk in rsys.substances])
    odesys, extra = create_odesys(rsys, rates_kw=dict(cstr_fr_fc=(fr, fc)))
    _check_cstr(odesys, fr, fc, extra_pars=dict(k2=5))


@requires("pygslodeiv2", "sym", units_library)
def test_get_odesys_rsys_with_units():
    rsys = ReactionSystem.from_string(
        """
    A -> B; 0.096/s
    B + C -> P; 4e3/M/s
    """,
        substance_factory=Substance,
    )
    with pytest.raises(Exception):
        get_odesys(
            rsys
        )  # not a strict test, SI_base_registry could be made the default

    odesys, extra = get_odesys(rsys, unit_registry=SI_base_registry)
    tend = 10
    tend_units = tend * u.s
    c0 = {"A": 1e-6, "B": 0, "C": 1, "P": 0}
    c0_units = {k: v * u.molar for k, v in c0.items()}
    result1 = odesys.integrate(tend_units, c0_units, integrator="gsl")
    assert result1.info["success"]

    with pytest.raises(Exception):
        odesys.integrate(tend, c0, integrator="gsl")


@requires("pyodeint", "sym", units_library)
def test_get_odesys_rsys_with_units__named_params():
    rsys = ReactionSystem.from_string(
        """
    A -> B; 'k1'
    B + C -> P; 'k2'
    """,
        substance_factory=Substance,
    )
    odesys, extra = get_odesys(
        rsys, include_params=False, unit_registry=SI_base_registry
    )
    tend = 10
    tend_units = tend * u.s
    c0 = {"A": 1e-6, "B": 0, "C": 1, "P": 0}
    p = {"k1": 3, "k2": 4}
    p_units = {"k1": 3 / u.s, "k2": 4 / u.M / u.s}
    c0_units = {k: v * u.molar for k, v in c0.items()}
    result1 = odesys.integrate(tend_units, c0_units, p_units, integrator="odeint")
    assert result1.info["success"]

    with pytest.raises(Exception):
        odesys.integrate(tend, c0, p, integrator="odeint")


@requires("pycvodes", "sym", units_library)
def test_get_odesys__Eyring():
    R = 8.314472
    T_K = 300
    dH = 80e3
    dS = 10
    rsys1 = ReactionSystem.from_string(
        """
    NOBr -> NO + Br; EyringParam(dH={dH}*J/mol, dS={dS}*J/K/mol)
    """.format(
            dH=dH, dS=dS
        ),
        substances="NOBr NO Br".split(),
    )
    kref = 20836643994.118652 * T_K * np.exp(-(dH - T_K * dS) / (R * T_K))
    NOBr0_M = 0.7
    init_cond = dict(NOBr=NOBr0_M * u.M, NO=0 * u.M, Br=0 * u.M)
    t = 5 * u.second
    params = dict(temperature=T_K * u.K)

    def check(rsys):
        odesys, extra = get_odesys(
            rsys, unit_registry=SI_base_registry, constants=const
        )
        res = odesys.integrate(t, init_cond, params, integrator="cvode")
        NOBr_ref = NOBr0_M * np.exp(-kref * to_unitless(res.xout, u.second))
        ref = np.array([NOBr_ref] + [NOBr0_M - NOBr_ref] * 2).T
        cmp = to_unitless(res.yout, u.M)
        assert np.allclose(cmp, ref)

    check(rsys1)
    rsys2 = ReactionSystem.from_string(
        """
    NOBr -> NO + Br; MassAction(EyringHS([{dH}*J/mol, {dS}*J/K/mol]))
    """.format(
            dH=dH, dS=dS
        ),
        substances="NOBr NO Br".split(),
    )
    check(rsys2)


@requires("pycvodes", "sym", units_library)
def test_get_odesys__Eyring_2nd_order():
    R = 8.314472
    T_K = 300
    dH = 80e3
    dS = 10
    rsys1b = ReactionSystem.from_string(
        """
    NO + Br -> NOBr; EyringParam(dH={dH}*J/mol, dS={dS}*J/K/mol)
    """.format(
            dH=dH, dS=dS
        )
    )
    c0 = 1  # mol/dm3 === 1000 mol/m3
    kbref = 20836643994.118652 * T_K * np.exp(-(dH - T_K * dS) / (R * T_K)) / c0
    NO0_M = 1.5
    Br0_M = 0.7
    init_cond = dict(NOBr=0 * u.M, NO=NO0_M * u.M, Br=Br0_M * u.M)
    t = 5 * u.second
    params = dict(temperature=T_K * u.K)

    def analytic_b(t):
        U, V = NO0_M, Br0_M
        d = U - V
        return (U * (1 - np.exp(-kbref * t * d))) / (U / V - np.exp(-kbref * t * d))

    def check(rsys):
        odesys, extra = get_odesys(
            rsys, unit_registry=SI_base_registry, constants=const
        )
        res = odesys.integrate(t, init_cond, params, integrator="cvode")
        t_sec = to_unitless(res.xout, u.second)
        NOBr_ref = analytic_b(t_sec)
        cmp = to_unitless(res.yout, u.M)
        ref = np.empty_like(cmp)
        ref[:, odesys.names.index("NOBr")] = NOBr_ref
        ref[:, odesys.names.index("Br")] = Br0_M - NOBr_ref
        ref[:, odesys.names.index("NO")] = NO0_M - NOBr_ref
        assert np.allclose(cmp, ref)

    check(rsys1b)
    rsys2b = ReactionSystem.from_string(
        """
    NO + Br -> NOBr; MassAction(EyringHS([{dH}*J/mol, {dS}*J/K/mol]))
    """.format(
            dH=dH, dS=dS
        )
    )
    check(rsys2b)


@requires("pycvodes", "sym", "scipy", units_library)
def test_get_odesys__Eyring_1st_order_linearly_ramped_temperature():
    from scipy.special import expi

    def analytic_unit0(t, T0, dH, dS):
        R = 8.314472
        kB = 1.3806504e-23
        h = 6.62606896e-34
        A = kB / h * np.exp(dS / R)
        B = dH / R
        return np.exp(
            A
            * (
                (-(B ** 2) * np.exp(B / T0) * expi(-B / T0) - T0 * (B - T0))
                * np.exp(-B / T0)
                + (
                    B ** 2 * np.exp(B / (t + T0)) * expi(-B / (t + T0))
                    - (t + T0) * (-B + t + T0)
                )
                * np.exp(-B / (t + T0))
            )
            / 2
        )

    T_K = 290
    dH = 80e3
    dS = 10
    rsys1 = ReactionSystem.from_string(
        """
    NOBr -> NO + Br; EyringParam(dH={dH}*J/mol, dS={dS}*J/K/mol)
    """.format(
            dH=dH, dS=dS
        )
    )

    NOBr0_M = 0.7
    init_cond = dict(NOBr=NOBr0_M * u.M, NO=0 * u.M, Br=0 * u.M)
    t = 20 * u.second

    def check(rsys):
        odes, extra = get_odesys(
            rsys,
            unit_registry=SI_base_registry,
            constants=const,
            substitutions={"temperature": RampedTemp([T_K * u.K, 1 * u.K / u.s])},
        )
        for odesys in [odes, odes.as_autonomous()]:
            res = odesys.integrate(t, init_cond, integrator="cvode")
            t_sec = to_unitless(res.xout, u.second)
            NOBr_ref = NOBr0_M * analytic_unit0(t_sec, T_K, dH, dS)
            cmp = to_unitless(res.yout, u.M)
            ref = np.empty_like(cmp)
            ref[:, odesys.names.index("NOBr")] = NOBr_ref
            ref[:, odesys.names.index("Br")] = NOBr0_M - NOBr_ref
            ref[:, odesys.names.index("NO")] = NOBr0_M - NOBr_ref
            assert np.allclose(cmp, ref)

    check(rsys1)
    rsys2 = ReactionSystem.from_string(
        """
    NOBr -> NO + Br; MassAction(EyringHS([{dH}*J/mol, {dS}*J/K/mol]))
    """.format(
            dH=dH, dS=dS
        )
    )
    check(rsys2)


@requires("pycvodes", "sym", "scipy", units_library)
def test_get_odesys__Eyring_2nd_order_linearly_ramped_temperature():
    from scipy.special import expi

    def analytic_unit0(t, k, m, dH, dS):
        R = 8.314472
        kB = 1.3806504e-23
        h = 6.62606896e-34
        A = kB / h * np.exp(dS / R)
        B = dH / R
        return (
            k
            * np.exp(B * (k * t + 2 * m) / (m * (k * t + m)))
            / (
                A
                * (
                    -(B ** 2) * np.exp(B / (k * t + m)) * expi(-B / (k * t + m))
                    - B * k * t
                    - B * m
                    + k ** 2 * t ** 2
                    + 2 * k * m * t
                    + m ** 2
                )
                * np.exp(B / m)
                + (
                    A * B ** 2 * np.exp(B / m) * expi(-B / m)
                    - A * m * (-B + m)
                    + k * np.exp(B / m)
                )
                * np.exp(B / (k * t + m))
            )
        )

    T_K = 290
    dTdt_Ks = 3
    dH = 80e3
    dS = 10
    rsys1 = ReactionSystem.from_string(
        """
    2 NO2 -> N2O4; EyringParam(dH={dH}*J/mol, dS={dS}*J/K/mol)
    """.format(
            dH=dH, dS=dS
        )
    )

    NO2_M = 1.0
    init_cond = dict(NO2=NO2_M * u.M, N2O4=0 * u.M)
    t = 20 * u.second

    def check(rsys):
        odes, extra = get_odesys(
            rsys,
            unit_registry=SI_base_registry,
            constants=const,
            substitutions={"temperature": RampedTemp([T_K * u.K, dTdt_Ks * u.K / u.s])},
        )
        for odesys in [odes, odes.as_autonomous()]:
            res = odesys.integrate(t, init_cond, integrator="cvode")
            t_sec = to_unitless(res.xout, u.second)
            NO2_ref = analytic_unit0(t_sec, dTdt_Ks, T_K, dH, dS)
            cmp = to_unitless(res.yout, u.M)
            ref = np.empty_like(cmp)
            ref[:, odesys.names.index("NO2")] = NO2_ref
            ref[:, odesys.names.index("N2O4")] = (NO2_M - NO2_ref) / 2
            assert np.allclose(cmp, ref)

    check(rsys1)
    rsys2 = ReactionSystem.from_string(
        """
    2 NO2 -> N2O4; MassAction(EyringHS([{dH}*J/mol, {dS}*J/K/mol]))
    """.format(
            dH=dH, dS=dS
        )
    )
    check(rsys2)


@requires("pycvodes", "sym", units_library)
def test_get_odesys__Eyring_2nd_order_reversible():
    R = 8.314472
    T_K = 273.15 + 20  # 20 degree celsius
    kB = 1.3806504e-23
    h = 6.62606896e-34

    dHf = 74e3
    dSf = R * np.log(h / kB / T_K * 1e16)

    dHb = 79e3
    dSb = dSf - 23

    rsys1 = ReactionSystem.from_string(
        """
    Fe+3 + SCN- -> FeSCN+2; EyringParam(dH={dHf}*J/mol, dS={dSf}*J/K/mol)
    FeSCN+2 -> Fe+3 + SCN-; EyringParam(dH={dHb}*J/mol, dS={dSb}*J/K/mol)
    """.format(
            dHf=dHf, dSf=dSf, dHb=dHb, dSb=dSb
        )
    )
    kf_ref = 20836643994.118652 * T_K * np.exp(-(dHf - T_K * dSf) / (R * T_K))
    kb_ref = 20836643994.118652 * T_K * np.exp(-(dHb - T_K * dSb) / (R * T_K))
    Fe0 = 6e-3
    SCN0 = 2e-3
    init_cond = {"Fe+3": Fe0 * u.M, "SCN-": SCN0 * u.M, "FeSCN+2": 0 * u.M}
    t = 3 * u.second

    def check(rsys, params):
        odes, extra = get_odesys(
            rsys, include_params=False, unit_registry=SI_base_registry, constants=const
        )
        for odesys in [odes, odes.as_autonomous()]:
            res = odesys.integrate(t, init_cond, params, integrator="cvode")
            t_sec = to_unitless(res.xout, u.second)
            FeSCN_ref = binary_rev(t_sec, kf_ref, kb_ref, 0, Fe0, SCN0)
            cmp = to_unitless(res.yout, u.M)
            ref = np.empty_like(cmp)
            ref[:, odesys.names.index("FeSCN+2")] = FeSCN_ref
            ref[:, odesys.names.index("Fe+3")] = Fe0 - FeSCN_ref
            ref[:, odesys.names.index("SCN-")] = SCN0 - FeSCN_ref
            assert np.allclose(cmp, ref)

    check(rsys1, {"temperature": T_K * u.K})
    rsys2 = ReactionSystem.from_string(
        """
    Fe+3 + SCN- -> FeSCN+2; MassAction(EyringHS([{dHf}*J/mol, {dSf}*J/K/mol]))
    FeSCN+2 -> Fe+3 + SCN-; MassAction(EyringHS([{dHb}*J/mol, {dSb}*J/K/mol]))
    """.format(
            dHf=dHf, dSf=dSf, dHb=dHb, dSb=dSb
        )
    )
    check(rsys2, {"temperature": T_K * u.K})
    rsys3 = ReactionSystem.from_string(
        """
    Fe+3 + SCN- -> FeSCN+2; MassAction(EyringHS.fk('dHf', 'dSf'))
    FeSCN+2 -> Fe+3 + SCN-; MassAction(EyringHS.fk('dHb', 'dSb'))
    """
    )
    check(
        rsys3,
        dict(
            temperature=T_K * u.K,
            dHf=dHf * u.J / u.mol,
            dSf=dSf * u.J / u.mol / u.K,
            dHb=dHb * u.J / u.mol,
            dSb=dSb * u.J / u.mol / u.K,
        ),
    )


@requires("numpy", "pyodesys", "sympy", "pycvodes")
def test_create_odesys():
    rsys = ReactionSystem.from_string(
        """
    A -> B; 'k1'
    B + C -> P; 'k2'
    """,
        substance_factory=Substance,
    )

    odesys, odesys_extra = create_odesys(rsys, unit_registry=SI_base_registry)

    tend_ul = 10
    init_conc_ul = {"A": 1e-6, "B": 0, "C": 1}
    params_ul = dict(k1=3, k2=4)

    tend = tend_ul * u.s
    params = {"k1": params_ul["k1"] / u.s, "k2": params_ul["k2"] / u.M / u.s}
    init_conc = {k: v * u.molar for k, v in init_conc_ul.items()}

    validation = odesys_extra["validate"](dict(init_conc, **params))
    (P,) = validation["not_seen"]
    assert P == "P"

    ref_rates = {
        "A": -params["k1"] * init_conc["A"],
        "P": params["k2"] * init_conc["B"] * init_conc["C"],
    }
    ref_rates["B"] = -ref_rates["A"] - ref_rates["P"]
    ref_rates["C"] = -ref_rates["P"]
    assert validation["rates"] == ref_rates

    result1, result1_extra = odesys_extra["unit_aware_solve"](
        tend, defaultdict(lambda: 0 * u.molar, init_conc), params, integrator="cvode"
    )
    assert result1.info["success"]

    result2 = odesys.integrate(
        tend_ul, defaultdict(lambda: 0, init_conc_ul), params_ul, integrator="cvode"
    )
    assert np.allclose(result2.yout[-1, :], to_unitless(result1.yout[-1, :], u.molar))


@requires("pycvodes", "sym", units_library)
def test_create_odesys__Radiolytic():
    rsys1 = ReactionSystem.from_string(
        """
    -> e-(aq); Radiolytic.fk('g_emaq')
    """,
        checks=(),
    )
    ic1 = {"e-(aq)": 0.0}
    t1 = 5
    p1 = dict(g_emaq=42.0, doserate=17.0, density=5.0)
    odesys1, odesys_extra = create_odesys(rsys1)
    result1 = odesys1.integrate(t1, ic1, p1)
    yref1 = result1.xout * p1["g_emaq"] * p1["doserate"] * p1["density"]
    assert np.allclose(yref1, result1.yout.squeeze())


@requires("pycvodes", "sym", units_library)
def test_create_odesys__validate__catalyst():
    rsys1 = ReactionSystem.from_string(
        """
    H2O2 + Pt -> 2 OH + Pt; 'k_decomp'
    """
    )
    ic1 = defaultdict(lambda: 0 * u.molar, {"H2O2": 3.0 * u.molar, "Pt": 0.5 * u.molar})
    t1 = linspace(0 * u.s, 0.3 * u.s, 7)
    p1 = dict(k_decomp=42 / u.second / u.molar)
    odesys1, odesys_extra = create_odesys(rsys1)
    validation = odesys_extra["validate"](dict(ic1, **p1))
    assert validation["not_seen"] == {"OH"}

    dedim_ctx = _mk_dedim(SI_base_registry)
    (t, c, _p), dedim_extra = dedim_ctx["dedim_tcp"](
        t1, [ic1[k] for k in odesys1.names], p1
    )
    result1 = odesys1.integrate(t, c, _p)
    tout = result1.xout * dedim_extra["unit_time"]
    cout = result1.yout * dedim_extra["unit_conc"]
    yref1 = ic1["H2O2"] * np.exp(-tout * ic1["Pt"] * p1["k_decomp"])
    assert allclose(yref1, cout[:, odesys1.names.index("H2O2")], rtol=1e-6)


@requires("pyodesys", units_library)
def test_create_odesys__ShiftedTPoly():
    rxn = Reaction({"A": 1, "B": 1}, {"C": 3, "D": 2}, "k_bi", {"A": 3})
    rsys = ReactionSystem([rxn], "A B C D")

    _k0, _k1, T0C = 10, 2, 273.15
    rate = MassAction(
        ShiftedTPoly([T0C * u.K, _k0 / u.molar / u.s, _k1 / u.molar / u.s / u.K])
    )
    T_C = 25
    T = (T0C + T_C) * u.kelvin
    p1 = rate.rate_coeff({"temperature": T})
    assert allclose(p1, (_k0 + _k1 * T_C) / u.molar / u.s)

    odesys, odesys_extra = create_odesys(rsys)
    ics = {"A": 3 * u.molar, "B": 5 * u.molar, "C": 11 * u.molar, "D": 13 * u.molar}
    pars = dict(k_bi=p1)
    validation = odesys_extra["validate"](dict(ics, **pars))
    assert set(map(str, validation["not_seen"])) == {"C", "D"}

    dedim_ctx = _mk_dedim(SI_base_registry)
    (t, c, _p), dedim_extra = dedim_ctx["dedim_tcp"](
        -37 * u.s, [ics[k] for k in odesys.names], pars
    )
    fout = odesys.f_cb(t, c, [_p[pk] for pk in odesys.param_names])
    r = 3 * 5 * (_k0 + _k1 * 25) * 1000  # mol/m3/s
    ref = [-4 * r, -r, 3 * r, 2 * r]
    assert np.all(abs((fout - ref) / ref) < 1e-14)

    odesys.integrate(t, c, _p)



# -*- coding: utf-8 -*-
"""
Module collecting classes and functions for dealing with (multiphase) chemical
equilibria.

.. Note::

  This module is provisional at the moment, i.e. the API is not stable and may
  break without a deprecation cycle.

"""

import warnings

import numpy as np

from .chemistry import equilibrium_quotient, Equilibrium, Species
from .reactionsystem import ReactionSystem
from ._util import get_backend
from .util.pyutil import deprecated
from ._eqsys import EqCalcResult, NumSysLin, NumSysLog, NumSysSquare as _NumSysSquare


NumSysSquare = deprecated()(_NumSysSquare)


class EqSystem(ReactionSystem):

    _BaseReaction = Equilibrium
    _BaseSubstance = Species

    def html(self, *args, **kwargs):
        k = "color_categories"
        kwargs[k] = kwargs.get(k, False)
        return super(EqSystem, self).html(*args, **kwargs)

    def eq_constants(self, non_precip_rids=(), eq_params=None, small=0):
        if eq_params is None:
            eq_params = [eq.param for eq in self.rxns]
        return [
            small if idx in non_precip_rids else eq for idx, eq in enumerate(eq_params)
        ]

    def equilibrium_quotients(self, concs):
        stoichs = self.stoichs()
        return [equilibrium_quotient(concs, stoichs[ri, :]) for ri in range(self.nr)]

    def stoichs_constants(
        self, eq_params=None, rref=False, Matrix=None, backend=None, non_precip_rids=()
    ):
        if eq_params is None:
            eq_params = self.eq_constants()
        if rref:
            from pyneqsys.symbolic import linear_rref

            be = get_backend(backend)
            rA, rb = linear_rref(
                self.stoichs(non_precip_rids), list(map(be.log, eq_params)), Matrix
            )
            return rA.tolist(), list(map(be.exp, rb))
        else:
            return (self.stoichs(non_precip_rids), eq_params)

    def composition_conservation(self, concs, init_concs):
        composition_vecs, comp_keys = self.composition_balance_vectors()
        A = np.array(composition_vecs)
        return (
            comp_keys,
            np.dot(A, self.as_per_substance_array(concs).T),
            np.dot(A, self.as_per_substance_array(init_concs).T),
        )

    def other_phase_species_idxs(self, phase_idx=0):
        return [
            idx
            for idx, s in enumerate(self.substances.values())
            if s.phase_idx != phase_idx
        ]

    @property
    @deprecated(
        last_supported_version="0.3.1",
        will_be_missing_in="0.8.0",
        use_instead=other_phase_species_idxs,
    )
    def precipitate_substance_idxs(self):
        return [idx for idx, s in enumerate(self.substances.values()) if s.precipitate]

    def phase_transfer_reaction_idxs(self, phase_idx=0):
        return [
            idx
            for idx, rxn in enumerate(self.rxns)
            if rxn.has_precipitates(self.substances)
        ]

    @property
    @deprecated(
        last_supported_version="0.3.1",
        will_be_missing_in="0.8.0",
        use_instead=phase_transfer_reaction_idxs,
    )
    def precipitate_rxn_idxs(self):
        return [
            idx
            for idx, rxn in enumerate(self.rxns)
            if rxn.has_precipitates(self.substances)
        ]

    def dissolved(self, concs):
        """Return dissolved concentrations"""
        new_concs = concs.copy()
        for r in self.rxns:
            if r.has_precipitates(self.substances):
                net_stoich = np.asarray(r.net_stoich(self.substances))
                s_net, s_stoich, s_idx = r.precipitate_stoich(self.substances)
                new_concs -= new_concs[s_idx] / s_stoich * net_stoich
        return new_concs

    def _fw_cond_factory(self, ri, rtol=1e-14):
        rxn = self.rxns[ri]

        def fw_cond(x, p):
            precip_stoich_coeff, precip_idx = rxn.precipitate_stoich(self.substances)[
                1:3
            ]
            q = rxn.Q(self.substances, self.dissolved(x))
            k = rxn.equilibrium_constant()
            if precip_stoich_coeff > 0:
                return q * (1 + rtol) < k
            elif precip_stoich_coeff < 0:
                return q > k * (1 + rtol)
            else:
                raise NotImplementedError

        return fw_cond

    def _bw_cond_factory(self, ri, small):
        rxn = self.rxns[ri]

        def bw_cond(x, p):
            precipitate_idx = rxn.precipitate_stoich(self.substances)[2]
            if x[precipitate_idx] < small:
                return False
            else:
                return True

        return bw_cond

    def _SymbolicSys_from_NumSys(
        self, NS, conds, rref_equil, rref_preserv, new_eq_params=True
    ):
        from pyneqsys.symbolic import SymbolicSys
        import sympy as sp

        ns = NS(
            self,
            backend=sp,
            rref_equil=rref_equil,
            rref_preserv=rref_preserv,
            precipitates=conds,
            new_eq_params=new_eq_params,
        )
        symb_kw = {}
        if ns.pre_processor is not None:
            symb_kw["pre_processors"] = [ns.pre_processor]
        if ns.post_processor is not None:
            symb_kw["post_processors"] = [ns.post_processor]
        if ns.internal_x0_cb is not None:
            symb_kw["internal_x0_cb"] = ns.internal_x0_cb
        return SymbolicSys.from_callback(
            ns.f,
            self.ns,
            nparams=self.ns + (self.nr if new_eq_params else 0),
            **symb_kw
        )

    def get_neqsys_conditional_chained(
        self, rref_equil=False, rref_preserv=False, NumSys=NumSysLin, **kwargs
    ):
        from pyneqsys import ConditionalNeqSys, ChainedNeqSys

        def factory(conds):
            return ChainedNeqSys(
                [
                    self._SymbolicSys_from_NumSys(
                        NS, conds, rref_equil, rref_preserv, **kwargs
                    )
                    for NS in NumSys
                ]
            )

        cond_cbs = [
            (self._fw_cond_factory(ri), self._bw_cond_factory(ri, NumSys[0].small))
            for ri in self.phase_transfer_reaction_idxs()
        ]
        return ConditionalNeqSys(cond_cbs, factory)

    def get_neqsys_chained_conditional(
        self, rref_equil=False, rref_preserv=False, NumSys=NumSysLin, **kwargs
    ):
        from pyneqsys import ConditionalNeqSys, ChainedNeqSys

        def mk_factory(NS):
            def factory(conds):
                return self._SymbolicSys_from_NumSys(
                    NS, conds, rref_equil, rref_preserv, **kwargs
                )

            return factory

        return ChainedNeqSys(
            [
                ConditionalNeqSys(
                    [
                        (self._fw_cond_factory(ri), self._bw_cond_factory(ri, NS.small))
                        for ri in self.phase_transfer_reaction_idxs()
                    ],
                    mk_factory(NS),
                )
                for NS in NumSys
            ]
        )

    def get_neqsys_static_conditions(
        self,
        rref_equil=False,
        rref_preserv=False,
        NumSys=(NumSysLin,),
        precipitates=None,
        **kwargs
    ):
        if precipitates is None:
            precipitates = (False,) * len(self.phase_transfer_reaction_idxs())
        from pyneqsys import ChainedNeqSys

        return ChainedNeqSys(
            [
                self._SymbolicSys_from_NumSys(
                    NS, precipitates, rref_equil, rref_preserv, **kwargs
                )
                for NS in NumSys
            ]
        )

    def get_neqsys(self, neqsys_type, NumSys=NumSysLin, **kwargs):
        new_kw = {"rref_equil": False, "rref_preserv": False}
        if neqsys_type == "static_conditions":
            new_kw["precipitates"] = None
        for k in new_kw:
            if k in kwargs:
                new_kw[k] = kwargs.pop(k)

        try:
            NumSys[0]
        except TypeError:
            new_kw["NumSys"] = (NumSys,)
        else:
            new_kw["NumSys"] = NumSys

        return getattr(self, "get_neqsys_" + neqsys_type)(**new_kw)

    def non_precip_rids(self, precipitates):
        return [
            idx
            for idx, precip in zip(self.phase_transfer_reaction_idxs(), precipitates)
            if not precip
        ]

    def _result_is_sane(self, init_concs, x, rtol=1e-9):
        sc_upper_bounds = np.array(self.upper_conc_bounds(init_concs))
        neg_conc, too_much = np.any(x < 0), np.any(x > sc_upper_bounds * (1 + rtol))
        if neg_conc or too_much:
            if neg_conc:
                warnings.warn("Negative concentration")
            if too_much:
                warnings.warn("Too much of at least one component")
            return False
        return True

    def _solve(
        self,
        init_concs,
        x0=None,
        NumSys=(NumSysLog, NumSysLin),
        neqsys="chained_conditional",
        **kwargs
    ):
        if isinstance(neqsys, str):
            neqsys = self.get_neqsys(
                neqsys,
                NumSys=NumSys,
                rref_equil=kwargs.pop("rref_equil", False),
                rref_preserv=kwargs.pop("rref_preserv", False),
                precipitates=kwargs.pop("precipitates", None),
            )
        if x0 is None:
            x0 = init_concs
        params = np.concatenate(
            (init_concs, [float(elem) for elem in self.eq_constants()])
        )
        x, sol = neqsys.solve(x0, params, **kwargs)
        if not sol["success"]:
            warnings.warn("Root-finding indicated as failed by solver.")
        sane = self._result_is_sane(init_concs, x)
        return x, sol, sane

    def solve(self, init_concs, varied=None, **kwargs):
        results = EqCalcResult(self, init_concs, varied)
        results.solve()
        return results

    def root(
        self,
        init_concs,
        x0=None,
        neqsys=None,
        NumSys=NumSysLog,
        neqsys_type="chained_conditional",
        **kwargs
    ):
        init_concs = self.as_per_substance_array(init_concs)
        params = np.concatenate(
            (init_concs, [float(elem) for elem in self.eq_constants()])
        )
        if neqsys is None:
            neqsys = self.get_neqsys(
                neqsys_type,
                NumSys=NumSys,
                rref_equil=kwargs.pop("rref_equil", False),
                rref_preserv=kwargs.pop("rref_preserv", False),
                precipitates=kwargs.pop("precipitates", None),
            )
        if x0 is None:
            x0 = init_concs
        x, sol = neqsys.solve(x0, params, **kwargs)
        if not sol["success"]:
            warnings.warn("Root finding indicated as failed by solver.")
        sane = self._result_is_sane(init_concs, x)
        return x, sol, sane

    @staticmethod
    def _get_default_plot_ax(subplot_kwargs=None):
        import matplotlib.pyplot as plt

        if subplot_kwargs is None:
            subplot_kwargs = dict(xscale="log", yscale="log")
        return plt.subplot(1, 1, 1, **subplot_kwargs)

    def substance_labels(self, latex=False):
        if latex:
            result = ["$" + s.latex_name + "$" for s in self.substances.values()]
            return result
        else:
            return [s.name for s in self.substances.values()]

    def roots(
        self,
        init_concs,
        varied_data,
        varied,
        x0=None,
        NumSys=NumSysLog,
        plot_kwargs=None,
        neqsys_type="chained_conditional",
        **kwargs
    ):
        """
        Parameters
        ----------
        init_concs : array or dict
        varied_data : array
        varied_idx : int or str
        x0 : array
        NumSys : _NumSys subclass
            See :class:`NumSysLin`, :class:`NumSysLog`, etc.
        plot_kwargs : dict
            See py:meth:`pyneqsys.NeqSys.solve`. Two additional keys
            are intercepted here:
                latex_names: bool (default: False)
                conc_unit_str: str (default: 'M')
        neqsys_type : str
            what method to use for NeqSys construction (get_neqsys_*)
        \\*\\*kwargs :
            Keyword arguments passed on to py:meth:`pyneqsys.NeqSys.solve_series`.

        """
        _plot = plot_kwargs is not None
        if _plot:
            latex_names = plot_kwargs.pop("latex_names", False)
            conc_unit_str = plot_kwargs.pop("conc_unit_str", "M")
            if "ax" not in plot_kwargs:
                plot_kwargs["ax"] = self._get_default_plot_ax()

        init_concs = self.as_per_substance_array(init_concs)
        neqsys = self.get_neqsys(
            neqsys_type,
            NumSys=NumSys,
            rref_equil=kwargs.pop("rref_equil", False),
            rref_preserv=kwargs.pop("rref_preserv", False),
            precipitates=kwargs.pop("precipitates", None),
        )
        if x0 is None:
            x0 = init_concs

        if _plot:
            cb = neqsys.solve_and_plot_series
            if "plot_kwargs" not in kwargs:
                kwargs["plot_kwargs"] = plot_kwargs
            if "labels" not in kwargs["plot_kwargs"]:
                kwargs["plot_kwargs"]["labels"] = self.substance_labels(latex_names)
            if "substances" in plot_kwargs:
                if "indices" in plot_kwargs:
                    raise ValueError("Now I am confused..")
                kwargs["plot_kwargs"]["indices"] = map(
                    self.as_substance_index, plot_kwargs.pop("substances")
                )
                print(kwargs["plot_kwargs"]["indices"])
        else:
            cb = neqsys.solve_series

        params = np.concatenate((init_concs, self.eq_constants()))
        xvecs, info_dicts = cb(
            x0,
            params,
            varied_data,
            self.as_substance_index(varied),
            propagate=False,
            **kwargs
        )
        sanity = [self._result_is_sane(init_concs, x) for x in xvecs]

        if _plot:
            import matplotlib.pyplot as plt
            from pyneqsys.plotting import mpl_outside_legend

            mpl_outside_legend(plt.gca())
            varied_subst = self.substances[varied]
            xlbl = (
                "$[" + varied_subst.latex_name + "]_0$"
                if latex_names
                else str(varied_subst)
            )
            plt.gca().set_xlabel(xlbl + " / " + conc_unit_str)
            plt.gca().set_ylabel("Concentration / " + conc_unit_str)

        return xvecs, info_dicts, sanity

    def plot_errors(
        self,
        concs,
        init_concs,
        varied_data,
        varied,
        axes=None,
        compositions=True,
        Q=True,
        subplot_kwargs=None,
    ):
        if axes is None:
            import matplotlib.pyplot as plt

            if subplot_kwargs is None:
                subplot_kwargs = dict(xscale="log")
            fig, axes = plt.subplots(1, 2, figsize=(10, 4), subplot_kw=subplot_kwargs)
        varied_idx = self.as_substance_index(varied)
        ls, c = "- -- : -.".split(), "krgbcmy"
        all_inits = np.tile(
            self.as_per_substance_array(init_concs), (len(varied_data), 1)
        )
        all_inits[:, varied_idx] = varied_data
        if compositions:
            cmp_nrs, m1, m2 = self.composition_conservation(concs, all_inits)
            for cidx, (cmp_nr, a1, a2) in enumerate(zip(cmp_nrs, m1, m2)):
                axes[0].plot(
                    concs[:, varied_idx],
                    a1 - a2,
                    label="Comp " + str(cmp_nr),
                    ls=ls[cidx % len(ls)],
                    c=c[cidx % len(c)],
                )
                axes[1].plot(
                    concs[:, varied_idx],
                    (a1 - a2) / np.abs(a2),
                    label="Comp " + str(cmp_nr),
                    ls=ls[cidx % len(ls)],
                    c=c[cidx % len(c)],
                )

        if Q:
            # TODO: handle precipitate phases in plotting Q error
            qs = self.equilibrium_quotients(concs)
            ks = [rxn.param for rxn in self.rxns]
            for idx, (q, k) in enumerate(zip(qs, ks)):
                axes[0].plot(
                    concs[:, varied_idx],
                    q - k,
                    label="K R:" + str(idx),
                    ls=ls[(idx + cidx) % len(ls)],
                    c=c[(idx + cidx) % len(c)],
                )
                axes[1].plot(
                    concs[:, varied_idx],
                    (q - k) / k,
                    label="K R:" + str(idx),
                    ls=ls[(idx + cidx) % len(ls)],
                    c=c[(idx + cidx) % len(c)],
                )

        from pyneqsys.plotting import mpl_outside_legend

        mpl_outside_legend(axes[0])
        mpl_outside_legend(axes[1])
        axes[0].set_title("Absolute errors")
        axes[1].set_title("Relative errors")



# -*- coding: utf-8 -*-
"""ChemPy is a Python package useful for solving problems in chemistry."""


from ._url import __url__
from ._release import __version__
from .chemistry import (
    Substance,
    Reaction,
    Equilibrium,
    Species,
    balance_stoichiometry,
    mass_fractions,
)
from .reactionsystem import ReactionSystem
from .henry import Henry
from .util.periodic import atomic_number
from .kinetics import EyringParam, EyringHS, MassAction

from .util.pyutil import ChemPyDeprecationWarning

from . import henry

import sys

if sys.version_info < (3, 5, 0):
    import warnings

    warnings.warn(
        "Use 'chempy<0.7' if using python versions < 3.5", ChemPyDeprecationWarning
    )



# -*- coding: utf-8 -*-

from .util.pyutil import ChemPyDeprecationWarning

import warnings

from .kinetics.arrhenius import (
    arrhenius_equation,
    fit_arrhenius_equation,
    ArrheniusParam,
    ArrheniusParamWithUnits,
)


warnings.warn("use .kinetics.arrhenius instead of .arrhenius", ChemPyDeprecationWarning)



# -*- coding: utf-8 -*-
"""
Module for dealing with constants `Henry's law
<https://en.wikipedia.org/wiki/Henry's_law>`_.
"""

from ._util import get_backend
from .util.pyutil import defaultnamedtuple, deprecated
from .units import default_units


def Henry_H_at_T(T, H, Tderiv, T0=None, units=None, backend=None):
    """Evaluate Henry's constant H at temperature T

    Parameters
    ----------
    T: float
        Temperature (with units), assumed to be in Kelvin if ``units == None``
    H: float
        Henry's constant
    Tderiv: float (optional)
        dln(H)/d(1/T), assumed to be in Kelvin if ``units == None``.
    T0: float
        Reference temperature, assumed to be in Kelvin if ``units == None``
        default: 298.15 K
    units: object (optional)
        object with attributes: kelvin (e.g. chempy.units.default_units)
    backend : module (optional)
        module with "exp", default: numpy, math

    """
    be = get_backend(backend)
    if units is None:
        K = 1
    else:
        K = units.Kelvin
    if T0 is None:
        T0 = 298.15 * K
    return H * be.exp(Tderiv * (1 / T - 1 / T0))


class Henry(defaultnamedtuple("Henry", "Hcp Tderiv T0 ref", [None, None])):
    """Henry's gas constant

    Note that the reference temperature
    is set by the attribute :py:attr:`T0` which defaults to
    298.15 (Kelvin).

    Parameters
    ----------
    Hcp: float
        Henry's constant [M/atm]
    Tderiv: float
        dln(kH)/d(1/T) [K]
        Equivalent to $\\Delta_{soln}H / R$
    ref: object
        Reference for origin of parameters
    units: object (optional)
        object with attributes: kelvin

    Examples
    --------
    >>> H_H2 = Henry(1.2e-3, 1800, ref='carpenter_1966')
    >>> '%.2g' % H_H2(298.15)
    '0.0012'

    """

    def __call__(self, T, units=None, backend=None):
        """Evaluates Henry's constant for provided temperature"""
        return Henry_H_at_T(
            T, self.Hcp, self.Tderiv, self.T0, units=units, backend=backend
        )

    @deprecated("0.3.1", "0.5.0", __call__)
    def get_kH_at_T(self, *args, **kwargs):
        return self(*args, **kwargs)

    def get_c_at_T_and_P(self, T, P, **kwargs):
        """Convenience method for calculating concentration

        Calculate what concentration is needed to achieve a given partial
        pressure at a specified temperature

        Parameters
        ----------
        T: float
            Temperature
        P: float
            Pressure
        \\*\\*kwargs:
            Keyword arguments passed on to :meth:`__call__`

        """
        return P * self(T, **kwargs)

    def get_P_at_T_and_c(self, T, c, **kwargs):
        """Convenience method for calculating concentration

        Calculate the partial pressure for given temperature and concentration


        Parameters
        ----------
        T: float
            Temperature
        P: float
            Pressure
        \\*\\*kwargs:
            Keyword arguments passed on to :meth:`__call__`
        """
        return c / self(T, **kwargs)


class HenryWithUnits(Henry):
    """Analogous to :class:`Henry`

    Examples
    --------
    >>> from chempy.units import to_unitless, default_units as u
    >>> H_CO = HenryWithUnits(9.7e-6 * u.mol/u.m**3/u.Pa, 1300*u.K, ref='sander_2015')
    >>> '%.2g' % to_unitless(H_CO(298.15 * u.K), u.molar/u.bar)
    '0.00097'

    """

    def __call__(self, T, units=default_units, backend=None):
        """Evaluates Henry's constant for provided temperature"""
        return super(HenryWithUnits, self).__call__(T, units, backend)



from chempy.util.testing import requires
from chempy.units import units_library, default_units as u
from ..gas_sol_electrolytes_schumpe_1993 import lg_solubility_ratio


@requires(units_library)
def test_lg_solubility_ratio():
    lgr = lg_solubility_ratio(
        {"Br-": 0.05 * u.molar, "Na+": 0.050 * u.molar}, "N2O", units=u
    )
    assert lgr != 0  # TODO: calculate by hand the reference value



# -*- coding: utf-8 -*-

import sympy
from ..salcs import (
    calc_salcs_projection,
    calc_salcs_func,
    _expand_irreducible,
    _angles_to_vectors,
)


def test_calc_salcs_projection():
    # ammonia hydrogens
    a, b, c = sympy.symbols('a b c')
    assert (calc_salcs_projection([a, b, c, a, b, c], 'c3v') ==
            [a + b + c, 0, a - b/2 - c/2])

    # trigonal bipyramidal
    a1, a2, e1, e2, e3 = sympy.symbols('a1, a2, e1, e2, e3')
    assert (calc_salcs_projection([e1, e2, e3, -e1, -e2, -e3, -e1,
                                   -e2, -e3, e1, e2, e3], 'd3h') ==
            [0, 0, 0, 0, e1 + e2 + e3, e1 - e2/2 - e3/2])
    assert (calc_salcs_projection([a1, a1, a1, -a2, -a2, -a2, -a2,
                                   -a2, -a2, a1, a1, a1], 'd3h') ==
            [a1 - a2, 0, 0, 0, a1 + a2, 0])

    # square planar s-orbitals
    a, b, c, d = sympy.symbols('a b c d')
    after_trans = [a, b, d, c, c, a, d, b, c, b, d, a, c, a, d, b]
    assert (calc_salcs_projection(after_trans, 'd4h') ==
            [a + b + c + d, 0, a - b + c - d, 0, 0, 0, 0, 0, 0, a - c])

    # benzene p-orbitals
    a, b, c, d, e, f = sympy.symbols('a b c d e f')
    after_trans = [a, b, f, c, e, d, -a, -c, -e, -b, -d, -f, -d, -c, -e, -b,
                   -f, -a, b, d, f, a, c, e]
    assert (calc_salcs_projection(after_trans, 'd6h') ==
            [0, 0, 0, a - b + c - d + e - f, a + b/2 - c/2 - d - e/2 + f/2,
             0, 0, a + b + c + d + e + f, 0, 0, 0,
             a - b/2 - c/2 + d - e/2 - f/2])

    # butadiene p-orbitals - inequivalent p-orbitals are treated separately
    a, b, c, d = sympy.symbols('a b c d')
    after_trans_outer = [a, -d, -a, d]
    after_trans_inner = [b, -c, -b, c]
    assert (calc_salcs_projection(after_trans_outer, 'c2v') ==
            [0, a - d, 0, a + d])
    assert (calc_salcs_projection(after_trans_inner, 'c2v') ==
            [0, b - c, 0, b + c])


def test_calc_salcs_func():
    # square planar
    a, b, c, d = sympy.symbols('a b c d')
    salc_true1 = [a + b + c + d, 0, a - b + c - d, 0, 0, 0, 0, 0, 0,
                  [a - c, b - d]]
    assert (calc_salcs_func([[1, 0, 0], [0, 1, 0], [-1, 0, 0], [0, -1, 0]],
                            'd4h', [a, b, c, d], mode='vector') == salc_true1)

    # trigonal bipyramidal - a is for axial and e is for equatorial
    a1, a2, e1, e2, e3 = sympy.symbols('a1, a2, e1, e2, e3')
    salc_true2 = [[e1 + e2 + e3, a1 + a2], 0,
                  [e1 - 0.5*e2 - 0.5*e3, e2 - e3, e1 - 0.5*e2 - 0.5*e3,
                   e2 - e3], 0, a1 - a2, 0]
    angles = [[0, 90], [120, 90], [240, 90], [0, 0], [0, 180]]
    assert (calc_salcs_func(angles, 'd3h', [e1, e2, e3, a1, a2], mode='angle')
           == salc_true2)

    # seesaw - such as SF4, a is for axial and e is for equatorial
    a1, a2, e1, e2 = sympy.symbols('a1 a2 e1 e2')
    salc_true3 = [[e1 + e2, a1 + a2, e1 + e2, e1 + e2], 0, a1 - a2,
                  [e1 - e2, e1 - e2]]
    assert (calc_salcs_func([[0, 90], [-180, 90], [90, 120], [-90, 120]],
                            'c2v', [a1, a2, e1, e2], mode='angle') ==
            salc_true3)

    # octahedral
    a, b, c, d, e, f = sympy.symbols('a b c d e f')
    salc_true4 = [a + b + c + d + e + f, 0,
                  [-0.5*a - 0.5*b - 0.5*c - 0.5*d + e + f, a - b + c - d],
                  0, 0, 0, 0, 0, [a - c, b - d, e - f], 0]
    oh_angle = calc_salcs_func([[0, 90], [90, 90], [180, 90], [270, 90],
                                [0, 0], [0, -180]], 'oh', [a, b, c, d, e, f],
                               mode='angle')
    oh_vector = calc_salcs_func([[1, 0, 0], [0, 1, 0], [-1, 0, 0], [0, -1, 0],
                                 [0, 0, 1], [0, 0, -1]], 'oh',
                                [a, b, c, d, e, f])
    assert oh_angle == salc_true4
    assert oh_vector == salc_true4


def test_expand_irreducible():
    assert _expand_irreducible([2, -1, 0], 'c3v') == [2, -1, -1, 0, 0, 0]


def test_angles_to_vectors():
    assert (_angles_to_vectors([[0, 90], [90, 90], [180, 90], [-90, 90]]) ==
            [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [-1.0, 0.0, 0.0],
             [0.0, -1.0, 0.0]])



# -*- coding: utf-8 -*-

from pkg_resources import parse_requirements, parse_version

import os
from operator import lt, le, eq, ne, ge, gt
import pytest

_relop = dict(zip("< <= == != >= >".split(), (lt, le, eq, ne, ge, gt)))


class requires(object):
    """Conditional skipping (on requirements) of tests in pytest

    Examples
    --------
    >>> @requires('numpy', 'scipy')
    ... def test_sqrt():
    ...     import numpy as np
    ...     assert np.sqrt(4) == 2
    ...     from scipy.special import zeta
    ...     assert zeta(2) < 2
    ...
    >>> @requires('numpy>=1.9.0')
    ... def test_nanmedian():
    ...     import numpy as np
    ...     a = np.array([[10.0, 7, 4], [3, 2, 1]])
    ...     a[0, 1] = np.nan
    ...     assert np.nanmedian(a) == 3
    ...

    """

    def __init__(self, *reqs):
        self.missing = []
        self.incomp = []
        self.requirements = list(parse_requirements(reqs))
        for req in self.requirements:
            try:
                mod = __import__(req.project_name)
            except ImportError:
                self.missing.append(req.project_name)
            else:
                try:
                    ver = parse_version(mod.__version__)
                except AttributeError:
                    pass
                else:
                    for rel, vstr in req.specs:
                        if not _relop[rel](ver, parse_version(vstr)):
                            self.incomp.append(str(req))

    def __call__(self, cb):
        r = "Unfulfilled requirements."
        if self.missing:
            r += " Missing modules: %s." % ", ".join(self.missing)
        if self.incomp:
            r += " Incomp versions: %s." % ", ".join(self.incomp)
        return skipif(self.missing or self.incomp, reason=r)(cb)


def skipif(predicate, *, reason):
    if os.environ.get("CHEMPY_SKIP_NO_TESTS", "0") == "1":
        return pytest.mark.skipif(False, reason=reason)
    else:
        return pytest.mark.skipif(predicate, reason=reason)



# -*- coding: utf-8 -*-

try:
    import numpy as np
except ImportError:
    np = None

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

from .. import ReactionSystem, Equilibrium
from ..units import get_derived_unit, to_unitless, default_units as u


def _dominant_reaction_effects(substance_key, rsys, rates, linthreshy, eqk1, eqk2, eqs):
    tot = np.zeros(rates.shape[0])
    reaction_effects = rsys.per_reaction_effect_on_substance(substance_key)
    data = []
    for ri, n in reaction_effects.items():
        tot += n * rates[..., ri]
        if ri in eqk1:
            otheri = eqk2[eqk1.index(ri)]
            y = n * rates[..., ri] + reaction_effects[otheri] * rates[..., otheri]
            rxn = eqs[eqk1.index(ri)]
        elif ri in eqk2:
            continue
        else:
            y = n * rates[..., ri]
            rxn = rsys.rxns[ri]
        if np.all(np.abs(y) < linthreshy):
            continue
        data.append((y, rxn))
    return data, tot


def _combine_rxns_to_eq(rsys):
    eqk1, eqk2 = zip(*rsys.identify_equilibria())
    eqs = [
        Equilibrium(
            rsys.rxns[i1].reac,
            rsys.rxns[i1].prod,
            (rsys.rxns[i1].param, rsys.rxns[i2].param),
            inact_reac=rsys.rxns[i1].inact_reac,
            inact_prod=rsys.rxns[i1].inact_prod,
        )
        for i1, i2 in zip(eqk1, eqk2)
    ]
    return eqk1, eqk2, eqs


def plot_reaction_contributions(
    xyp,
    rsys,
    rate_exprs_cb,
    substance_keys=None,
    varied=None,
    axes=None,
    total=False,
    linthreshy=1e-9,
    relative=False,
    xscale="log",
    yscale="symlog",
    xlabel="Time",
    ylabel=None,
    combine_equilibria=False,
    selection=slice(None),
    unit_registry=None,
):
    """Plots per reaction contributions to concentration evolution of a substance.

    Parameters
    ----------
    xyp : ``pyodesys.results.Result`` instance or length 3 tuple or xout,yout,params
    result : pyodesys.results.Result
    substance_key : str

    """
    from pyodesys.results import Result

    if isinstance(xyp, Result):
        xyp = xyp.odesys.to_arrays(xyp.xout, xyp.yout, xyp.params, reshape=False)
    if varied is None:
        varied = xyp[0]
    if xyp[1].shape[-2] != varied.size:
        raise ValueError("Size mismatch between varied and yout")
    if substance_keys is None:
        substance_keys = rsys.substances.keys()
    if axes is None:
        _fig, axes = plt.subplots(len(substance_keys))
    rates = rate_exprs_cb(*xyp)
    if unit_registry is not None:
        time_unit = get_derived_unit(unit_registry, "time")
        conc_unit = get_derived_unit(unit_registry, "concentration")
        rates = to_unitless(rates * conc_unit / time_unit, u.molar / u.second)

    eqk1, eqk2, eqs = _combine_rxns_to_eq(rsys) if combine_equilibria else ([], [], [])

    for sk, ax in zip(substance_keys, axes):
        data, tot = _dominant_reaction_effects(
            sk, rsys, rates, linthreshy, eqk1, eqk2, eqs
        )
        factor = 1 / xyp[1][:, rsys.as_substance_index(sk)] if relative else 1
        if total:
            ax.plot(varied, factor * tot, c="k", label="Total", linewidth=2, ls=":")
        for y, rxn in sorted(data, key=lambda args: args[0][-1], reverse=True):
            ax.plot(
                varied, factor * y, label=r"$\mathrm{%s}$" % rxn.latex(rsys.substances)
            )

        if rsys.substances[sk].latex_name is None:
            ttl = rsys.substances[sk].name
            ttl_template = "%s"
        else:
            ttl = rsys.substances[sk].latex_name
            ttl_template = r"\mathrm{$%s$}"

        if yscale == "symlog":
            ax.axhline(linthreshy, linestyle="--", color="k", linewidth=0.5)
            ax.axhline(-linthreshy, linestyle="--", color="k", linewidth=0.5)
            ax.set_yscale(yscale, linthresh=linthreshy)
        else:
            ax.set_yscale(yscale)

        if ylabel is None:
            ax.set_ylabel(r"$\frac{d}{dt}\left[%s\right]\ /\ M\cdot s^{-1}$" % ttl)
        else:
            ax.set_ylabel(ylabel)
            ax.set_title(ttl_template % ttl)

        ax.set_xlabel(xlabel)
        ax.set_xscale(xscale)
        ax.legend(loc="best")


def dominant_reactions_graph(
    concs,
    rate_exprs_cb,
    rsys,
    substance_key,
    linthreshy=1e-9,
    fname="dominant_reactions_graph.png",
    relative=False,
    combine_equilibria=False,
    **kwargs
):
    from ..util.graph import rsys2graph

    rates = rate_exprs_cb(0, concs)
    eqk1, eqk2, eqs = _combine_rxns_to_eq(rsys) if combine_equilibria else ([], [], [])
    rrate, rxns = zip(
        *_dominant_reaction_effects(
            substance_key, rsys, rates, linthreshy, eqk1, eqk2, eqs
        )[0]
    )
    rsys = ReactionSystem(rxns, rsys.substances, rsys.name)
    lg_rrate = np.log10(np.abs(rrate))
    rsys2graph(rsys, fname=fname, penwidths=1 + lg_rrate - np.min(lg_rrate), **kwargs)



from chempy.util.testing import requires
from ..integrated import pseudo_irrev, pseudo_rev, binary_irrev, binary_rev


try:
    import sympy
except ImportError:
    sympy = None
else:
    one = sympy.S(1)

    t, kf, kb, prod, major, minor = sympy.symbols(
        "t kf kb prod major minor", negative=False, nonnegative=True, real=True
    )

    subsd = {
        t: one * 2,
        kf: one * 3,
        kb: one * 7,
        major: one * 11,
        minor: one * 13,
        prod: one * 0,
    }


@requires("sympy")
def test_pseudo_irrev():
    f = pseudo_irrev(t, kf, prod, major, minor, backend=sympy)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    assert (num_dfdt - (major * kf * (minor - f)).subs(subsd)).simplify() == 0


@requires("sympy")
def test_pseudo_rev():
    f = pseudo_rev(t, kf, kb, prod, major, minor, backend=sympy)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    assert (num_dfdt - (major * kf * (minor - f) - kb * f).subs(subsd)).simplify() == 0


@requires("sympy")
def test_binary_irrev():
    f = binary_irrev(t, kf, prod, major, minor, backend=sympy)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    assert (num_dfdt - (kf * (minor - f) * (major - f)).subs(subsd)).simplify() == 0


@requires("sympy")
def test_binary_rev():
    f = binary_rev(t, kf, kb, prod, major, minor, backend=sympy)
    dfdt = f.diff(t)
    num_dfdt = dfdt.subs(subsd)
    ans = kf * (minor - f) * (major - f) - kb * f
    # symbolic susbsitution fails:
    assert abs(float(num_dfdt) - float(ans.subs(subsd))) < 2e-14



# -*- coding: utf-8 -*-
""" The units module provides the following attributes:

- ``chempy.units.default_units``
- ``chempy.units.default_constants``
- ``chempy.units.SI_base_registry``

together with some functions.

Currently `quantities <https://pypi.python.org/pypi/quantities>`_ is used as
the underlying package to handle units. If it is possible you should try to
only use the ``chempy.units`` module (since it is likely that ``ChemPy``
will change this backend at some point in the future). Therefore you should not
rely on any attributes of the ``Quantity`` instances (and rather use
getter & setter functions in `chempy.units`).

"""

from functools import reduce
from operator import mul
import sys
import warnings
from typing import Iterable

from .util.arithmeticdict import ArithmeticDict
from .util.pyutil import NameSpace, deprecated

units_library = "quantities"  # info used for selective testing.


try:
    pq = __import__(units_library)
except ImportError:
    UncertainQuantity = None
    default_constants = None
    default_units = None
    SI_base_registry = None
    np = None
else:
    import numpy as np
    from .util._quantities import _patch_quantities

    _patch_quantities(pq)
    UncertainQuantity = pq.UncertainQuantity
    # Let us extend the underlying pq namespace with some common units in
    # chemistry
    default_constants = NameSpace(pq.constants)

    default_units = NameSpace(pq)
    default_units.dm = default_units.decimetre = pq.UnitQuantity(
        "decimetre", default_units.m / 10.0, u_symbol="dm"
    )
    default_units.m3 = default_units.metre ** 3
    default_units.dm3 = default_units.decimetre ** 3
    default_units.cm3 = default_units.centimetre ** 3
    if not hasattr(default_units, "molar"):
        default_units.molar = pq.UnitQuantity(
            "M", 1e3 * default_units.mole / default_units.m3, u_symbol="M"
        )
    if not hasattr(default_units, "millimolar"):
        default_units.millimolar = pq.UnitQuantity(
            "mM", 1 * default_units.mole / default_units.m3, u_symbol="mM"
        )
    if not hasattr(default_units, "micromolar"):
        default_units.micromolar = pq.UnitQuantity(
            "uM", 1e-3 * default_units.mole / default_units.m3, u_symbol="μM"
        )
    if not hasattr(default_units, "nanomolar"):
        default_units.nanomolar = pq.UnitQuantity(
            "nM", 1e-6 * default_units.mole / default_units.m3, u_symbol="nM"
        )
    if not hasattr(default_units, "molal"):
        default_units.molal = pq.UnitQuantity(
            "molal", default_units.mole / default_units.kg, u_symbol="molal"
        )
    if not hasattr(default_units, "per100eV"):
        default_units.per100eV = pq.UnitQuantity(
            "per100eV",
            1 / (100 * default_units.eV * default_constants.Avogadro_constant),
            u_symbol="(100eV)**-1",
        )
    if not hasattr(default_units, "micromole"):
        default_units.micromole = pq.UnitQuantity(
            "micromole", pq.mole / 1e6, u_symbol="μmol"
        )
    if not hasattr(default_units, "nanomole"):
        default_units.nanomole = pq.UnitQuantity(
            "nanomole", pq.mole / 1e9, u_symbol="nmol"
        )
    if not hasattr(default_units, "kilojoule"):
        default_units.kilojoule = pq.UnitQuantity(
            "kilojoule", 1e3 * pq.joule, u_symbol="kJ"
        )
    if not hasattr(default_units, "kilogray"):
        default_units.kilogray = pq.UnitQuantity(
            "kilogray", 1e3 * pq.gray, u_symbol="kGy"
        )
    if not hasattr(default_units, "perMolar_perSecond"):
        default_units.perMolar_perSecond = 1 / default_units.molar / pq.s
    if not hasattr(default_units, "umol"):
        default_units.umol = default_units.micromole
    if not hasattr(default_units, "umol_per_J"):
        default_units.umol_per_J = default_units.umol / pq.joule

    # unit registry data and logic:

    SI_base_registry = {
        "length": default_units.metre,
        "mass": default_units.kilogram,
        "time": default_units.second,
        "current": default_units.ampere,
        "temperature": default_units.kelvin,
        "luminous_intensity": default_units.candela,
        "amount": default_units.mole,
    }


def magnitude(value):
    try:
        return value.magnitude
    except AttributeError:
        return value


def uncertainty(uval):
    uncert = uval.uncertainty
    if not is_quantity(uncert):
        warnings.warn(f"Handling unexpected type: {type(uval)}")
    return uncert


def simplified(value):
    if hasattr(value, "simplified"):
        return value.simplified
    else:
        return to_unitless(value)


def is_quantity(arg):
    if arg.__class__.__name__ == "Quantity":
        return True  # this checks works even if quantities is not installed.
    else:
        return False


# SI Base Quantities:
time = ArithmeticDict(int, {"time": 1})
length = ArithmeticDict(int, {"length": 1})
mass = ArithmeticDict(int, {"mass": 1})
current = ArithmeticDict(int, {"current": 1})
temperature = ArithmeticDict(int, {"temperature": 1})
amount = ArithmeticDict(int, {"amount": 1})
# intensity = ArithmeticDict(int, {'intensity': 1}) what's wrong with photon flux? (human eyes, bah!)

energy = ArithmeticDict(int, {"mass": 1, "length": 2, "time": -2})
volume = ArithmeticDict(int, {"length": 3})
concentration = {"amount": 1} - volume


def get_derived_unit(registry, key):
    """Get the unit of a physical quantity in a provided unit system.

    Parameters
    ----------
    registry: dict (str: unit)
        mapping 'length', 'mass', 'time', 'current', 'temperature',
        'luminous_intensity', 'amount'. If registry is ``None`` the
        function returns 1.0 unconditionally.
    key: str
        one of the registry keys or one of: 'diffusivity', 'electricalmobility',
        'permittivity', 'charge', 'energy', 'concentration', 'density',
        'radiolytic_yield'.

    Examples
    --------
    >>> m, s = default_units.meter, default_units.second
    >>> bool(get_derived_unit(SI_base_registry, 'diffusivity') == m**2/s)
    True

    """
    if registry is None:
        return 1.0
    derived = {
        "diffusivity": registry["length"] ** 2 / registry["time"],
        "electrical_mobility": (
            registry["current"] * registry["time"] ** 2 / registry["mass"]
        ),
        "permittivity": (
            registry["current"] ** 2
            * registry["time"] ** 4
            / (registry["length"] ** 3 * registry["mass"])
        ),
        "charge": registry["current"] * registry["time"],
        "energy": registry["mass"] * registry["length"] ** 2 / registry["time"] ** 2,
        "concentration": registry["amount"] / registry["length"] ** 3,
        "density": registry["mass"] / registry["length"] ** 3,
    }
    derived["diffusion"] = derived["diffusivity"]  # 'diffusion' is deprecated
    derived["radiolytic_yield"] = registry["amount"] / derived["energy"]
    derived["doserate"] = derived["energy"] / registry["mass"] / registry["time"]
    derived["linear_energy_transfer"] = derived["energy"] / registry["length"]

    try:
        return derived[key]
    except KeyError:
        return registry[key]


def unit_registry_to_human_readable(unit_registry):
    """Serialization of a unit registry."""
    if unit_registry is None:
        return None
    new_registry = {}
    integer_one = 1
    for k in SI_base_registry:
        if unit_registry[k] is integer_one:
            new_registry[k] = 1, 1
        else:
            dim_list = list(unit_registry[k].dimensionality)
            if len(dim_list) != 1:
                raise TypeError("Compound units not allowed: {}".format(dim_list))
            u_symbol = dim_list[0].u_symbol
            new_registry[k] = float(unit_registry[k]), u_symbol
    return new_registry


def _latex_from_dimensionality(dim):
    # see https://github.com/python-quantities/python-quantities/issues/148
    from quantities.markup import format_units_latex

    return format_units_latex(dim, mult=r"\\cdot")


def latex_of_unit(quant):
    """Returns LaTeX representation of the unit of a quantity

    Examples
    --------
    >>> print(latex_of_unit(1/default_units.kelvin))
    \\mathrm{\\frac{1}{K}}

    """
    return _latex_from_dimensionality(quant.dimensionality).strip("$")


def unicode_of_unit(quant):
    """Returns unicode representation of the unit of a quantity

    Examples
    --------
    >>> print(unicode_of_unit(1/default_units.kelvin))
    1/K

    """
    return quant.dimensionality.unicode


def html_of_unit(quant):
    """Returns HTML representation of the unit of a quantity

    Examples
    --------
    >>> print(html_of_unit(2*default_units.m**2))
    m<sup>2</sup>

    """
    return quant.dimensionality.html


def unit_registry_from_human_readable(unit_registry):
    """Deserialization of unit_registry."""
    if unit_registry is None:
        return None
    new_registry = {}
    for k in SI_base_registry:
        factor, u_symbol = unit_registry[k]
        if u_symbol == 1:
            unit_quants = [1]
        else:
            unit_quants = list(pq.Quantity(0, u_symbol).dimensionality.keys())

        if len(unit_quants) != 1:
            raise TypeError("Unknown UnitQuantity: {}".format(unit_registry[k]))
        else:
            new_registry[k] = factor * unit_quants[0]
    return new_registry


# Abstraction of underlying package providing units and dimensional analysis:


def is_unitless(expr):
    """Returns ``True`` if ``expr`` is unitless, otherwise ``False``

    Examples
    --------
    >>> is_unitless(42)
    True
    >>> is_unitless(42*default_units.kilogram)
    False

    """
    if hasattr(expr, "dimensionality"):
        if expr.dimensionality == pq.dimensionless:
            return True
        else:
            return expr.simplified.dimensionality == pq.dimensionless.dimensionality
    if isinstance(expr, dict):
        return all(is_unitless(_) for _ in expr.values())
    elif isinstance(expr, (tuple, list)):
        return all(is_unitless(_) for _ in expr)
    return True


def unit_of(expr, simplified=False):
    """Returns the unit of a quantity

    Examples
    --------
    >>> bool(unit_of(42*pq.second) == unit_of(12*pq.second))
    True
    >>> unit_of(42)
    1

    """
    if isinstance(expr, (tuple, list)):
        return unit_of(uniform(expr)[0], simplified)
    elif isinstance(expr, dict):
        return unit_of(list(uniform(expr).values())[0], simplified)

    try:
        if simplified:
            return expr.units.simplified
        else:
            return expr.units
    except AttributeError:
        return 1


def rescale(value, unit):
    try:
        return value.rescale(unit)
    except AttributeError:
        if unit == 1:
            return value
        else:
            raise


def to_unitless(value, new_unit=None):
    """ Nondimensionalization of a quantity.

    Parameters
    ----------
    value: quantity
    new_unit: unit

    Examples
    --------
    >>> '%.1g' % to_unitless(1*default_units.metre, default_units.nm)
    '1e+09'
    >>> '%.1g %.1g' % tuple(to_unitless([1*default_units.m, 1*default_units.mm], default_units.nm))
    '1e+09 1e+06'

    """
    integer_one = 1
    if new_unit is None:
        new_unit = pq.dimensionless

    if isinstance(value, np.ndarray) and not hasattr(value, 'rescale'):
        if is_unitless(new_unit) and new_unit == 1:
            if value.ndim == 0:
                return value.item()
            else:
                return value
        return np.array([to_unitless(elem, new_unit) for elem in value])
    elif isinstance(value, dict):
        new_value = dict(value.items())  # value.copy()
        for k in value:
            new_value[k] = to_unitless(value[k], new_unit)
        return new_value
    elif isinstance(value, (int, float)) and new_unit is integer_one or new_unit is None:
        return value
    elif isinstance(value, Iterable) and getattr(value, 'ndim', -1) != 0:
        if len(value) == 0:
            return to_unitless(value[0], new_unit)
        else:
            return np.array([to_unitless(elem, new_unit) for elem in value])
    elif isinstance(value, str):
        raise ValueError("str not supported")

    ori_unit = unit_of(value)
    ori_mag = magnitude(value)
    scale_factor = simplified(ori_unit/new_unit)
    if is_unitless(scale_factor):
        sc = magnitude(scale_factor)
        if getattr(ori_mag, 'ndim', -1) == 0 or not isinstance(ori_mag, Iterable):
            return ori_mag*sc
        else:
            return np.array(ori_mag*sc)
    else:
        raise ValueError("Incompatible units")


def uniform(container):
    """Turns a list, tuple or dict with mixed units into one with uniform units.

    Parameters
    ----------
    container : tuple, list or dict

    Examples
    --------
    >>> km, m = default_units.kilometre, default_units.metre
    >>> uniform(dict(a=3*km, b=200*m))  # doctest: +SKIP
    {'b': array(200.0) * m, 'a': array(3000.0) * m}

    """
    if isinstance(container, (tuple, list)):
        unit = unit_of(container[0])
    elif isinstance(container, dict):
        unit = unit_of(list(container.values())[0])
        return container.__class__(
            [(k, to_unitless(v, unit) * unit) for k, v in container.items()]
        )
    else:
        return container
    return to_unitless(container, unit) * unit


def get_physical_dimensionality(value):
    if is_unitless(value):
        return {}
    _quantities_mapping = {
        pq.UnitLength: "length",
        pq.UnitMass: "mass",
        pq.UnitTime: "time",
        pq.UnitCurrent: "current",
        pq.UnitTemperature: "temperature",
        pq.UnitLuminousIntensity: "luminous_intensity",
        pq.UnitSubstance: "amount",
    }
    return {
        _quantities_mapping[k.__class__]: v
        for k, v in uniform(value).simplified.dimensionality.items()
    }


@deprecated(use_instead=get_physical_dimensionality, will_be_missing_in="0.8.0")
def get_physical_quantity(value):
    return get_physical_dimensionality(value)


def _get_unit_from_registry(dimensionality, registry):
    return reduce(mul, [registry[k] ** v for k, v in dimensionality.items()])


def default_unit_in_registry(value, registry):
    _dimensionality = get_physical_dimensionality(value)
    if _dimensionality == {}:
        return 1
    return _get_unit_from_registry(_dimensionality, registry)


def unitless_in_registry(value, registry):
    _default_unit = default_unit_in_registry(value, registry)
    return to_unitless(value, _default_unit)


# NumPy like functions for compatibility:


def compare_equality(a, b):
    """Returns True if two arguments are equal.

    Both arguments need to have the same dimensionality.

    Parameters
    ----------
    a : quantity
    b : quantity

    Examples
    --------
    >>> km, m = default_units.kilometre, default_units.metre
    >>> compare_equality(3*km, 3)
    False
    >>> bool(compare_equality(3*km, 3000*m))
    True

    """
    # Work around for https://github.com/python-quantities/python-quantities/issues/146
    try:
        a + b
    except TypeError:
        # We might be dealing with e.g. None (None + None raises TypeError)
        try:
            len(a)
        except TypeError:
            # Assumed scalar
            return a == b
        else:
            if len(a) != len(b):
                return False
            return all(compare_equality(_a, _b) for _a, _b in zip(a, b))
    except ValueError:
        return False
    else:
        return a == b


def allclose(a, b, rtol=1e-8, atol=None):
    """Analogous to ``numpy.allclose``."""
    if a.__class__.__name__ == "UncertainQuantity":
        return allclose(pq.Quantity(a), b, rtol=rtol, atol=atol)
    if b.__class__.__name__ == "UncertainQuantity":
        return allclose(a, pq.Quantity(b), rtol=rtol, atol=atol)

    try:
        d = abs(a - b)
    except Exception:
        try:
            if len(a) == len(b):
                return all(allclose(_a, _b, rtol, atol) for _a, _b in zip(a, b))
            else:
                return False
        except Exception:
            return False
    lim = abs(a) * rtol
    if atol is not None:
        lim += atol

    try:
        len(d)
    except TypeError:
        return d <= lim
    else:
        try:
            len(lim)
        except TypeError:
            return np.all([_d <= lim for _d in d])
        else:
            return np.all([_d <= _lim for _d, _lim in zip(d, lim)])


def linspace(start, stop, num=50):
    """Analogous to ``numpy.linspace``.

    Examples
    --------
    >>> bool(abs(linspace(2, 8, num=3)[1] - 5) < 1e-15)
    True

    """

    # work around for quantities v0.10.1 and NumPy
    unit = unit_of(start)
    start_ = to_unitless(start, unit)
    stop_ = to_unitless(stop, unit)
    return np.linspace(start_, stop_, num) * unit


def logspace_from_lin(start, stop, num=50):
    """Logarithmically spaced data points

    Examples
    --------
    >>> bool(abs(logspace_from_lin(2, 8, num=3)[1] - 4) < 1e-15)
    True

    """
    unit = unit_of(start)
    start_ = np.log2(to_unitless(start, unit))
    stop_ = np.log2(to_unitless(stop, unit))
    return np.exp2(np.linspace(start_, stop_, num)) * unit


def _sum(iterable):
    try:
        result = next(iterable)
    except TypeError:
        result = iterable[0]
        for elem in iterable[1:]:
            result += elem
        return result
    else:
        try:
            while True:
                result += next(iterable)
        except StopIteration:
            return result
        else:
            raise ValueError("Not sure how this point was reached")


class Backend(object):
    """Wrapper around modules such as numpy and math

    Instances of Backend wraps a module, e.g. `numpy` and ensures that
    arguments passed on are unitless, i.e. it raises an error if a
    transcendental function is used with quantities with units.

    Parameters
    ----------
    underlying_backend : module, str or tuple of str
        e.g. 'numpy' or ('sympy', 'math')

    Examples
    --------
    >>> import math
    >>> km, m = default_units.kilometre, default_units.metre
    >>> math.exp(3*km) == math.exp(3*m)
    True
    >>> be = Backend('math')
    >>> be.exp(3*km)  # DOCTEST: +ELLIPSIS
    Traceback (most recent call last):
        ...
    ValueError: Incompatible units
    >>> import numpy as np
    >>> np.sum([1000*pq.metre/pq.kilometre, 1]).item()
    1001.0
    >>> be_np = Backend(np)
    >>> be_np.sum([[1000*pq.metre/pq.kilometre, 1], [3, 4]], axis=1)
    array([2., 7.])

    """

    def __init__(self, underlying_backend=("numpy", "math")):
        if isinstance(underlying_backend, tuple):
            for name in underlying_backend:
                try:
                    self.be = __import__(name)
                except ImportError:
                    continue
                else:
                    break
            else:
                raise ValueError("Could not import any of %s" % str(underlying_backend))
        elif isinstance(underlying_backend, str):
            self.be = __import__(underlying_backend)
        else:
            self.be = underlying_backend

    def __getattr__(self, attr):
        be_attr = getattr(self.be, attr)
        if callable(be_attr):
            return lambda *args, **kwargs: be_attr(*map(to_unitless, args), **kwargs)
        else:
            return be_attr


# TODO: decide whether to deprecate in favor of "number_to_scientific_latex"?
def format_string(value, precision="%.5g", tex=False):
    """Formats a scalar with unit as two strings

    Parameters
    ----------
    value: float with unit
    precision: str
    tex: bool
       LaTeX formatted or not? (no '$' signs)

    Examples
    --------
    >>> print(' '.join(format_string(0.42*default_units.mol/default_units.decimetre**3)))
    0.42 mol/decimetre**3
    >>> print(' '.join(format_string(2/default_units.s, tex=True)))
    2 \\mathrm{\\frac{1}{s}}

    """
    if tex:
        unit_str = latex_of_unit(value)
    else:
        from quantities.markup import config

        attr = "unicode" if config.use_unicode else "string"
        unit_str = getattr(value.dimensionality, attr)
    return precision % float(value.magnitude), unit_str


def concatenate(arrays, **kwargs):
    """Patched version of numpy.concatenate

    Examples
    --------
    >>> from chempy.units import default_units as u
    >>> all(concatenate(([2, 3]*u.s, [4, 5]*u.s)) == [2, 3, 4, 5]*u.s)
    True

    """
    unit = unit_of(arrays[0])
    result = np.concatenate([to_unitless(arr, unit) for arr in arrays], **kwargs)
    return result * unit


def tile(array, *args, **kwargs):
    """Patched version of numpy.tile (with support for units)"""
    try:
        elem = array[0, ...]
    except TypeError:
        elem = array[0]

    unit = unit_of(elem)
    result = np.tile(to_unitless(array, unit), *args, **kwargs)
    return result * unit


def polyfit(x, y, deg, **kwargs):
    u_x = unit_of(x[0])
    u_y = unit_of(y[0])
    _x, _y = to_unitless(x, u_x), to_unitless(y, u_y)
    p = np.polyfit(_x, _y, deg)
    return [v * u_y * u_x ** (i - deg) for i, v in enumerate(p)]


def polyval(p, x):
    try:
        u_x = unit_of(x[0])
    except (TypeError, IndexError):
        u_x = unit_of(x)
    u_y = unit_of(p[-1])
    deg = len(p) - 1
    _p = [to_unitless(v, u_y * u_x ** (i - deg)) for i, v in enumerate(p)]
    _x = to_unitless(x, u_x)
    _y = np.polyval(_p, _x)
    return _y * u_y


def _wrap_numpy(k):
    numpy_func = getattr(np, k)
    if sys.version_info[0] > 2:
        from functools import wraps
    else:

        def wraps(_meta_fun):
            return lambda x: x  # py2: numpy.ufunc lacks "__module__"

    @wraps(numpy_func)
    def f(*args, **kwargs):
        return numpy_func(*map(to_unitless, args), **kwargs)

    return f


if np is None:
    patched_numpy = None
else:
    patched_numpy = NameSpace(np)
    patched_numpy.allclose = allclose
    patched_numpy.concatenate = concatenate
    patched_numpy.linspace = linspace
    patched_numpy.tile = tile
    patched_numpy.polyfit = polyfit
    patched_numpy.polyval = polyval
    for k in "log log10 log2 log1p exp expm1 logaddexp logaddexp2".split():
        setattr(patched_numpy, k, _wrap_numpy(k))


def fold_constants(arg):
    if hasattr(arg, "dimensionality"):
        m = arg.magnitude
        d = 1
        for k, v in arg.dimensionality.items():
            if isinstance(k, pq.UnitConstant):
                m = m * k.simplified ** v
            else:
                d = d * k ** v
        return m * d
    else:
        return arg



# -*- coding: utf-8 -*-

"""
Contains chemical group theory functions for calculating numbers and type of
irreducible representations in a reducible representation and number of IR
and Raman active modes. Below uses the inverse matrix method of solving for
irreducible representations described in J. Chem. Educ. 2009, 86, 251-253.
https://doi.org/10.1021/ed086p251.
"""

import numpy as np
from functools import wraps
from tabulate import tabulate

# data tables in character_tables.py file
from .tables import (
    tables,
    rot_trans_modes,
    IR_active,
    Raman_active,
    masks,
    atom_contribution,
    mulliken,
    headers,
    row_coeffs,
    column_coeffs
)


def print_header(group):
    """
    Print the header for a character table given the point group.

    Parameters
    ----------
    group : str
        Point group Schoenflies notation (e.g., 'C2v').  This is
        case-insensitive.

    Returns
    -------
    Prints character table header indicating order of symmetry operations.
    """
    symbols = headers[group.lower()]
    numbers = column_coeffs[group.lower()]

    header = []
    for i in range(len(numbers)):
        if numbers[i] != 1:
            header.append(str(numbers[i]) + symbols[i])
        else:
            header.append(symbols[i])

    print(*header)


def print_point_groups():
    """
    Print supported point group Schoenflies notations.

    Parameters
    ----------
    None

    Returns
    -------
    Prints Schoenflies notations for character tables available.

    """
    pg = ('C1', 'Cs', 'Ci', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'D2',
          'D3', 'D4', 'D5', 'D6', 'C2v', 'C3v', 'C4v', 'C5v', 'C6v', 'C2h',
          'C3h', 'C4h', 'C5h', 'C6h', 'D2h', 'D3h', 'D4h', 'D5h', 'D6h', 'D8h',
          'D2d', 'D3d', 'D4d', 'D5d', 'D6d', 'S4', 'S6', 'S8', 'T', 'Th', 'Td',
          'O', 'Oh', 'I', 'Ih')

    print(*pg)


def print_mulliken(group):
    """
    Print Mulliken symbols of irreducible representation in order.

    Parameters
    ----------
    group : str
        Point group Schoenflies notation (e.g., 'C2v').  This is
        case-insensitive.

    Returns
    -------
    Prints Mulliken notations for irreducible representations

    """
    print(*mulliken[group.lower()])


def print_table(group):
    """
    Print character table for given point group.

    Parameters
    ----------
    group : str
        Point group Schoenflies notation (e.g., 'C2v').  This is
        case-insensitive.

    Returns
    -------
    Prints the character table for the point group.

    """
    group = group.lower()
    symbols = headers[group]
    mull_symbols = mulliken[group]
    numbers = column_coeffs[group.lower()]

    # generate list of symmetry symbols with coefficients
    header = []
    for i in range(len(numbers)):
        if numbers[i] != 1:
            header.append(str(numbers[i]) + symbols[i])
        else:
            header.append(symbols[i])

    # generate list of Mulliken symbols including duplicates if imaginary rep
    if group.lower() in row_coeffs.keys():
        rows = []
        for i in range(len(mull_symbols)):
            rows.extend([mull_symbols[i]] * row_coeffs[group][i])
    else:
        rows = mull_symbols

    table = [[group.capitalize(), *header]]
    if group == 'c1':
        table.append([rows, 1])
    else:
        for i_row in range(len(rows)):
            table.append([rows[i_row], *tables[group][i_row]])

    print(tabulate(table, tablefmt='rounded_grid'))


@np.vectorize
def sympy_to_num(sp):
    """
    Convert array with sympy objects to array of floats and ints.

    Convert numpy array of sympy objects from table dictionary in tables.py
    to numpy array of floats and ints for use in calculations. Three types of
    values occur in the tables: ints, real sympy values, and imaginary sympy
    values.

    Parameters
    ----------
    sp : numpy array with sympy objects
        numpy array with sympy objects from table dictionary in tables.py

    Returns
    -------
    numpy array of floats and ints.

    """
    try:
        # converts sympy real values to floats
        return sp.evalf()
    except AttributeError:
        try:
            # convert sympy complex values to python complex numbers
            return complex(sp)
        except TypeError:
            # returns ints unchanged
            return sp


class Reducible:
    """Reducible representation object.

    Reducible representation object for calculating number and type of
    irreducible representations and calculating IR and Raman active vibrational
    modes.
    """

    def __init__(self, gamma, group, all_motion=False):
        """
        Initialize Reducible representation object.

        Parameters
        ----------
        gamma : List, tuple, or array
            Reducible representation.
        group : str
            Point group Schoenflies notation (e.g., 'C2v').  This is
            case-insensitive.
        all_motion : bool, optional
            False if reducible representation is only vibrational modes and
            True if the reducible describes all motions (rotational,
            translational, and vibrational). The default is False.

        Returns
        -------
        None.
        """
        if np.all(np.mod(gamma, 1) != 0):
            print('Invalid representation - must be whole numbers.')
        elif group.lower() not in tables.keys():
            print('Invalid point group.')
        elif np.array(gamma).size != tables[group.lower()].shape[0]:
            print(f'Invalid representation size for {group} point group.')
        else:
            self.group = group.lower()
            self.gamma = gamma
            self.all_motion = all_motion

    def return_dict_method(func):
        """
        Return results as a dictionary.

        Return a list or array as a dictionary with Mulliken symbols as
        the keys.

        Parameters
        ----------
        arr : List, tuple, or array
            List or array containing results corresponding to irreducible
            representations.
        group : str
            Point group Schoenflies notation (e.g., 'C2v').  This is
            case-insensitive.

        Returns
        -------
        Dictionary.

        """
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if kwargs.get('to_dict'):
                keys = mulliken[self.group.lower()]
                values = func(self, *args, **kwargs)
                return dict(zip(keys, values.tolist()))
            else:
                return func(self, *args, **kwargs)
        return wrapper

    @return_dict_method
    def decomp(self, to_dict=False):
        """
        Decompose reducible representation into number of irreducibles.

        Decompose a reducible representation for a specific point group and
        return the number of each irreducible representation in the reducible.

        The order of irreducibles can be looked up by print_mulliken()
        or print_table().

        Parameters
        ----------
        to_dict : bool
            True causes the function to return a dictionary with Mulliken
            symbols as the keys.

        Returns
        -------
        NumPy array with the number of each irreducible representation in the
        provided reducible representation. Use print_mulliken() for listing of
        irreducible representations and order. If to_dict=True, the function
        returns a dictionary with the Mulliken symbols as keys.

        Examples
        --------
        >>> rep = Reducible([9, -1, 3, 1], 'c2v', all_motion=True)
        >>> rep.decomp()
        array([3, 1, 3, 2])
        >>> rep = Reducible([15, 0, 0, 7, -2, -2], 'C3h', all_motion=False)
        >>> rep.decomp()
        array([3, 4, 2, 1])
        >>> rep.decomp(to_dict=True)
        {"A'": 3, "E'": 4, 'A"': 2, 'E"': 1}
        """
        table = sympy_to_num(tables[self.group])
        gamma = np.array(self.gamma)

        if self.group == 'c1':
            n_i = self.gamma
        else:
            mask = np.array(masks[self.group], dtype=bool)
            # mask removes complex conjugate to avoid "doubling problem"
            n_i = gamma.dot(np.linalg.inv(table)).real[mask]

        return np.rint(n_i).astype(int)

    @return_dict_method
    def vibe_modes(self, to_dict=False):
        """Return vibrational modes.

        Return the number of vibrational modes after rotation and translation
        are subtracted out.

        Parameters
        ----------
        to_dict : bool
            True causes the function to return a dictionary with Mulliken
            symbols as the keys.

        Returns
        -------
        Numpy array or dictionary

        Examples
        --------
        >>> rep = Reducible([9, -1, 3, 1], 'c2v', all_motion=True)
        >>> rep.vibe_modes()
        array([2, 0, 1, 0])
        >>> rep.vibe_modes(to_dict=True)
        {'A1': 2, 'A2': 0, 'B1': 1, 'B2': 0}
        """
        if self.all_motion is False:
            return self.decomp()
        else:
            rot_trans = rot_trans_modes[self.group]
            irreducibles = self.decomp()

        return np.array(irreducibles) - np.array(rot_trans)

    @return_dict_method
    def ir_active(self, to_dict=False):
        """Return IR active vibrational modes.

        Return the number of each irreducible representation that is IR active
        in the given reducible representation. If all_motion=True for the
        Reducible, the rotational and translational modes are automatically
        subtracted out.

        Parameters
        ----------
        to_dict : bool
            True causes function to return a dictionary with Mulliken symbols
            as the keys.

        Returns
        -------
        Numpy array or dictionary

        Examples
        --------
        >>> rep = Reducible([9, -1, 3, 1], 'c2v', all_motion=True)
        >>> rep.ir_active()
        array([2, 0, 1, 0])
        >>> rep = Reducible([5, 2, 1, 3, 0, 3], 'd3h', all_motion=False)
        >>> rep.ir_active()
        array([0, 0, 1, 0, 1, 0])
        >>> rep.ir_active(to_dict=True)
        {"A'1": 0, "A'2": 0, "E'": 1, 'A"1': 0, 'A"2': 1, 'E"': 0}

        """
        return self.vibe_modes() * np.array(IR_active[self.group])

    @return_dict_method
    def raman_active(self, to_dict=False):
        """Return Raman active vibrational modes.

        Return the number of each irreducible representation that is Raman
        active in the given reducible representation. If all_motion=True for
        the Reducible, the rotational and translational modes are automatically
        subtracted out.

        Parameters
        ----------
        to_dict : bool
            True causes the function to return a dictionary with Mulliken
            symbols as the keys.

        Returns
        -------
        Numpy array or dictionary

        Examples
        --------
        >>> rep = Reducible([9, -1, 3, 1], 'c2v', all_motion=True)
        >>> rep.raman_active([3, 1, 3, 2, 'C2v'])
        array([2, 0, 1, 0])
        >>> rep = Reducible([5, 2, 1, 3, 0, 3], 'd3h', all_motion=False)
        >>> rep.raman_active()
        array([2, 0, 1, 0, 0, 0])
        >>> rep.raman_active(to_dict=True)
        {"A'1": 2, "A'2": 0, "E'": 1, 'A"1': 0, 'A"2': 0, 'E"': 0}

        """
        return self.vibe_modes() * np.array(Raman_active[self.group])

    @classmethod
    def from_irred(cls, n_irred, group, all_motion=False):
        """Create reducible from number of irreducible representations.

        Alternative constructor that returns a Reducible representation object
        given the number of each irreducible representation that comprises
        the reducible representation in the point group.

        Parameters
        ----------
        n_irred: List, tuple, or array
            Number of each irreducible representation in the returned
            reducible representation.
        group: str
            Point group in Schoenflies notation (e.g., 'C2v').  This is
            case-insensitive.
        all_motion: bool
            Whether the resulting reducible representation represents all
            motions (rotation, vibration, and translation).

        Returns
        -------
        Reducible object

        Examples
        --------
        >>> rep = Reducible.from_irred([1, 0, 1, 0], 'c2v')
        >>> rep.gamma
        array([2, 0, 2, 0])
        >>> rep = Reducible.from_irred([3, 1, 1], 'C3v')
        >>> rep.gamma
        array([6, 3, 2])
        """
        irred_sum = np.sum((tables[group.lower()].T * n_irred).T, axis=0)

        return cls(np.rint(irred_sum).astype(int), group, all_motion=False)

    @classmethod
    def from_atoms(cls, n_atoms, group):
        """Create a representation based on number of stationary atoms.

        Alternative constructor that returns a Reducible representation object
        with all motions (rotation, translation, and vibration) given the
        number of atoms that do NOT move (i.e., translate) when carrying out
        each symmetry operation in the point group.

        Note: all_motion parameter in the returned Reducible object is set
        to True.

        Parameters
        ----------
        n_atoms: List, tuple, or array
            Number of atoms that remain stationary during each operation in
            the point group - see print_header() for the group operations.
        group: str
            Point group in Schoenflies notation (e.g., 'C2v').  This is
            case-insensitive.

        Returns
        -------
        Reducible object

        Examples
        --------
        >>> rep = Reducible.from_atoms([4, 2, 4, 2], 'c2v')
        >>> rep.gamma  # doctest: +NORMALIZE_WHITESPACE
        array([12, -2, 4, 2])

        """
        n_atoms = np.array(n_atoms)

        if np.all(np.mod(n_atoms, 1) == 0):
            gamma = np.rint(n_atoms * np.array(atom_contribution[group]))
            return cls(gamma.astype(int), group, all_motion=True)
        else:
            print("""Number of stationary atoms (n_atoms) must be an integer
                  value.""")



# -*- coding: utf-8 -*-

from ..testing import requires
from ..numutil import broadcast_stack


@requires("numpy")
def test_broadcast_stack():
    import numpy as np

    a = np.arange(9.0)
    b = broadcast_stack(np.arange(5.0), np.arange(5.0, 9.0))
    assert a.shape == b.shape and np.all(a == b)

    c = broadcast_stack([0, 1, 2], [3], 4, as_scalars=True)
    d = np.array([[0, 3, 4], [1, 3, 4], [2, 3, 4]])
    assert c.shape == d.shape and np.all(c == d)



import numpy as np
from chempy.tests.ammonical_cupric_solution import get_ammonical_cupric_eqsys


class TimeEqsys:
    def setup(self):
        self.eqsys, self.c0 = get_ammonical_cupric_eqsys()

    def time_roots(self):
        x, new_inits, success = self.eqsys.roots(self.c0, np.logspace(-3, 0, 50), "NH3")
        assert all(success)

    def time_roots_symengine(self):
        from symengine import Lambdify

        x, new_inits, success = self.eqsys.roots(
            self.c0,
            np.logspace(-3, 0, 50),
            "NH3",
            lambdify=Lambdify,
            lambdify_unpack=False,
        )
        assert all(success)

    def time_roots_no_propagate(self):
        x, new_inits, success = self.eqsys.roots(
            self.c0, np.logspace(-3, 0, 50), "NH3", propagate=False
        )
        assert all(success)


if __name__ == "__main__":
    import time

    te = TimeEqsys()
    te.setup()

    # t1 = time.time()
    # te.time_roots_symengine()
    # print(time.time()-t1)

    t1 = time.time()
    te.time_roots()
    print(time.time() - t1)



# -*- coding: utf-8 -*-
"""
This module implements an expression for estimating how gas solubility in water
is affected by dissolved non-reactive salts, the expression and parameters are from Schumpe (1993).
"""

import warnings

Tref_K = 298.2  # Kelvin

p_ion_rM = {  # "rM" === "per molar"
    "H+": 0.0,
    "Li+": 0.0691,
    "Na+": 0.1171,
    "K+": 0.0959,
    "Rb+": 0.0845,
    "Cs+": 0.0660,
    "NH4+": 0.0539,
    "Mg+2": 0.1765,
    "Ca+2": 0.1771,
    "Ba+2": 0.2021,
    "Fe+2": 0.1712,
    "Co+2": 0.1983,
    "Ni+2": 0.2039,
    "Cu+2": 0.1810,
    "Mn+2": 0.1620,
    "Zn+2": 0.1712,
    "Cd+2": 0.2201,
    "Al+3": 0.2253,
    "Fe+3": 0.0996,
    "Cr+3": 0.0595,
    "OH-": 0.0756,
    "F-": 0.1016,
    "Cl-": 0.0334,
    "Br-": 0.0137,
    "I-": 0.0020,
    "NO3-": 0.0050,
    "ClO4-": 0.0502,
    "IO4-": 0.1514,
    "HCO3-": 0.1372,
    "HSO3-": 0.0543,
    "H2PO4-": 0.1025,
    # '0-O-': 0.0765,
    # '0-OCH2COO-': 0.0119,
    "S2O3-2": 0.1109,
    "HPO4-2": 0.1789,
    "CO3-2": 0.1666,
    "SO3-2": 0.1537,
    "SO4-2": 0.1185,
    "PO4-3": 0.2117,
}

p_gas_rM = {
    "O2": 0.0,
    "CO2": -0.0183,
    "N2O": -0.0110,
    "C2H2": -0.0174,
    "C2H4": 0.0014,
    "He": -0.036,
    "Ne": -0.020,
    "Ar": -0.009,
    "Kr": 0.003,
    "Xe": 0.005,
    "Rn": 0.015,
    "H2": -0.024,
    "N2": -0.008,
    "NO": 0.004,
    "C2H6": 0.011,
}


def lg_solubility_ratio(electrolytes, gas, units=None, warn=True):
    """Returns the log10 value of the solubilty ratio

    Implements equation 16, p 156. from Schumpe (1993)

    Parameters
    ----------
    electrolytes : dict
        Mapping substance key (one in ``p_ion_rM``) to concentration.
    gas : str
        Substance key for the gas (one in ``p_gas_rM``).
    units : object (optional)
        object with attribute: molar
    warn : bool (default: True)
        Emit UserWarning when 'F-' among electrolytes.

    """
    if units is None:
        M = 1
    else:
        M = units.molar
    if warn and "F-" in electrolytes:
        warnings.warn("In Schumpe 1993: data for fluoride uncertain.")
    return sum(
        [(p_gas_rM[gas] / M + p_ion_rM[k] / M) * v for k, v in electrolytes.items()]
    )


reference = dict(
    doi="10.1016/0009-2509(93)80291-W",
    url="http://www.sciencedirect.com/science/article/pii/000925099380291W",
    year=1993,
    publisher="Pergamon",
    author="Schumpe, Adrian",
    title="The estimation of gas solubilities in salt solutions",
    volume=48,
    issn="0009-2509",
    abstract=(
        "The effects of dissolved salts on the solubilities of gases were analysed based on "
        "a comprehensive set of literature data for the temperature of 298.2 K. A new "
        "empirical model was suggested which, at no increase in the number of adjustable "
        "parameters, described the data with a lower standard deviation than previously "
        "suggested models. The parameter values evaluated for the new model allow to "
        "estimate the effects of 20 cations and 19 anions on the solubilities of 15 gases."
    ),
    pages=(153, 158),
    number=1,
    journaltitle="Chemical Engineering Science",
    shortjournal="Chemical Engineering Science",
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
This script generates a "coverage" badge as a svg file from
the html report from coverage.py

Usage:

   $ ./coverage_badge.py htmlcov/ coverage.svg

"""

import os

# this template was generated from shields.io on 2015-10-11
template = """
<svg xmlns="http://www.w3.org/2000/svg" width="92" height="20">
<linearGradient id="b" x2="0" y2="100%">
<stop offset="0" stop-color="#bbb" stop-opacity=".1"/>
<stop offset="1" stop-opacity=".1"/>
</linearGradient>
<mask id="a">
<rect width="92" height="20" rx="3" fill="#fff"/>
</mask>
<g mask="url(#a)">
<path fill="#555" d="M0 0h63v20H0z"/>
<path fill="{0:s}" d="M63 0h29v20H63z"/>
<path fill="url(#b)" d="M0 0h92v20H0z"/>
</g>
<g fill="#fff" text-anchor="middle" font-family="DejaVu Sans,Verdana,Geneva,
    sans-serif" font-size="11">
<text x="31.5" y="15" fill="#010101" fill-opacity=".3">coverage</text>
<text x="31.5" y="14">coverage</text>
<text x="76.5" y="15" fill="#010101" fill-opacity=".3">{1:s}%</text>
<text x="76.5" y="14">{1:s}%</text>
</g>
</svg>
"""


def get_coverage(htmldir):
    for line in open(os.path.join(htmldir, "index.html"), "rt"):
        if "pc_cov" in line:
            return int(line.split("pc_cov")[1].split(">")[1].split("<")[0].rstrip("%"))
    raise ValueError("Could not find pc_cov in index.html")


def write_cov_badge_svg(path, percent):
    colors = "#e05d44 #fe7d37 #dfb317 #a4a61d #97CA00 #4c1".split()
    limits_le = 50, 60, 70, 80, 90, 100
    c = next(clr for lim, clr in zip(limits_le, colors) if percent <= lim)
    with open(path, "wt") as f:
        f.write(template.format(c, str(percent)))


if __name__ == "__main__":
    import sys

    assert len(sys.argv) == 3
    cov_percent = get_coverage(sys.argv[1])
    write_cov_badge_svg(sys.argv[2], cov_percent)



# -*- coding: utf-8 -*-

import collections

import pytest

try:
    import numpy as np
except ImportError:
    np, NumSysLin, NumSysLog = [None] * 3
else:
    from ..equilibria import EqSystem, NumSysLin, NumSysLog
    from .ammonical_cupric_solution import get_ammonical_cupric_eqsys

from ..util.testing import requires
from ..chemistry import Substance, Reaction, Equilibrium, Species


@requires("numpy")
def test_EqSystem():
    a, b = sbstncs = Substance("a"), Substance("b")
    rxns = [Reaction({"a": 1}, {"b": 1})]
    es = EqSystem(rxns, collections.OrderedDict([(s.name, s) for s in sbstncs]))
    assert es.net_stoichs().tolist() == [[-1, 1]]


def _get_es1():
    a, b = sbstncs = Species("a"), Species("b")
    rxns = [Equilibrium({"a": 1}, {"b": 1}, 3)]
    return EqSystem(rxns, sbstncs)


def _get_es_water(EqSys=None):
    if EqSys is None:
        EqSys = EqSystem
    H2O = Substance("H2O", charge=0, composition={1: 2, 8: 1})
    OHm = Substance("OH-", charge=-1, composition={1: 1, 8: 1})
    Hp = Substance("H+", charge=1, composition={1: 1})
    Kw = 1e-14 / 55.5
    w_auto_p = Equilibrium({"H2O": 1}, {"Hp": 1, "OHm": 1}, Kw)
    return EqSys([w_auto_p], [H2O, OHm, Hp])


@requires("numpy")
def test_EqSystem_1():
    es = _get_es1()
    assert es.stoichs().tolist() == [[-1, 1]]
    assert es.eq_constants() == [3]


@requires("numpy")
def test_Equilibria_arithmetics():
    es1 = _get_es1()
    (e,) = es1.rxns
    e2 = 2 * e
    sum2 = e + e
    assert sum2 == e2


@requires("numpy")
def test_Equilibria_root():
    eqsys, c0 = get_ammonical_cupric_eqsys()
    x, sol, sane = eqsys.root(c0, NumSys=(NumSysLog,))
    assert sol["success"] and sane


def _species(Cls):
    return (
        Cls("H2O", 0, composition={1: 2, 8: 1}),
        Cls("H+", 1, composition={1: 1}),
        Cls("OH-", -1, composition={1: 1, 8: 1}),
        Cls("NH4+", 1, composition={1: 4, 7: 1}),
        Cls("NH3", 0, composition={1: 3, 7: 1}),
    )


@requires("numpy")
def test_Equilibria_root_simple():
    species = water, hydronium, hydroxide, ammonium, ammonia = _species(Species)

    water_auto_protolysis = Equilibrium(
        {water.name: 1}, {hydronium.name: 1, hydroxide.name: 1}, 1e-14 / 55.5
    )
    ammonia_protolysis = Equilibrium(
        {ammonium.name: 1}, {hydronium.name: 1, ammonia.name: 1}, 10 ** -9.26 / 55.5
    )
    eqsys = EqSystem([water_auto_protolysis, ammonia_protolysis], species)

    init_concs = collections.defaultdict(float, {water.name: 55.5, ammonia.name: 1e-3})
    x, sol1, sane1 = eqsys.root(init_concs)
    x, sol2, sane2 = eqsys.root(init_concs, x, NumSys=(NumSysLog,))
    assert sane2
    ref = eqsys.as_per_substance_array(
        {
            water.name: 55.5,
            ammonia.name: 1e-3 - 6.2e-4,
            ammonium.name: 6.2e-4,
            hydronium.name: 1.6e-11,
            hydroxide.name: 6.2e-4,
        }
    )
    assert np.allclose(x, ref, rtol=0.02, atol=1e-16)


def _get_NaCl(Cls, **precipitate_kwargs):
    Na_p, Cl_m, NaCl = sbstncs = (
        Cls("Na+", 1, composition={11: 1}),
        Cls("Cl-", -1, composition={17: 1}),
        Cls("NaCl", composition={11: 1, 17: 1}, **precipitate_kwargs),
    )
    sp = Equilibrium({"NaCl": 1}, {"Na+": 1, "Cl-": 1}, 4.0)
    eqsys = EqSystem([sp], sbstncs)
    cases = [
        [(0, 0, 0.1), (0.1, 0.1, 0)],
        [(0.5, 0.5, 0.4), (0.9, 0.9, 0)],
        [(1, 1, 1), (2, 2, 0)],
        [(0, 0, 2), (2, 2, 0)],
        [(3, 3, 3), (2, 2, 4)],
        [(3, 3, 0), (2, 2, 1)],
        [(0, 0, 3), (2, 2, 1)],
    ]
    return eqsys, [s.name for s in sbstncs], cases


@requires("numpy")
def test_EqSystem_dissolved():
    eqsys, names, _ = _get_NaCl(Cls=Species, phase_idx=1)
    inp = eqsys.as_per_substance_array({"Na+": 1, "Cl-": 2, "NaCl": 4})
    result = eqsys.dissolved(inp)
    ref = eqsys.as_per_substance_array({"Na+": 5, "Cl-": 6, "NaCl": 0})
    assert np.allclose(result, ref)


@pytest.mark.xfail  # TODO: fails in github-actions, passes locally (on 2 different machines..)
@requires("numpy")
@pytest.mark.parametrize("NumSys", [(NumSysLin,), (NumSysLog,), (NumSysLog, NumSysLin)])
def test_precipitate(NumSys):
    eqsys, species, cases = _get_NaCl(Species, phase_idx=1)

    for init, final in cases:
        x, sol, sane = eqsys.root(
            dict(zip(species, init)), NumSys=NumSys, rref_preserv=True, tol=1e-12
        )
        assert sol["success"] and sane
        assert x is not None
        assert np.allclose(x, np.asarray(final))
